import pandas as pd

from tdda import rexpy

df = pd.DataFrame({'a3': ["one", "two", pd.np.NaN],
                   'a45': ['three', 'four', 'five']})

re3 = rexpy.pdextract(df['a3'])
re45 = rexpy.pdextract(df['a45'])
re345 = rexpy.pdextract([df['a3'], df['a45']])

print('  re3: %s' % re3)
print(' re45: %s' % re45)
print('re345: %s' % re345)







from __future__ import print_function

from tdda.rexpy import extract
from tdda.rexpy.seq import common_string_sequence
from tdda.rexpy.relib import re

x = extract(['Roger', 'Coger', 'Doger'], tag=True, as_object=True)
print(x)

patternToExamples = x.pattern_matches()

sequences = []
for j, (pattern, examples) in enumerate(patternToExamples.items()):
    N = len(examples)
    if N < 1:
        print('%s:%s' % (pattern, examples))
    else:
        eparts = [re.match(x.results.rex[j], e).groups() for e in examples]
        nparts = len(eparts[0])
        for i in range(nparts):
            (L, R) = (eparts[0][i], eparts[1][i])
            n = 2
            s = common_string_sequence(L, R)
            while n < N and s != '':
                s = common_string_sequence(s, eparts[n][i])
                n += 1

            sequences.append(s)
print(sequences)




# -*- coding: utf-8 -*-

"""
URLs1

 http\:\/\/[a-z]{3,4}\.[a-z]{3,4}
 http\:\/\/[a-z]{5,19}\.com\/
 [a-z]{4,5}\:\/\/[a-z]{3}\.[a-z]{6,19}\.com
 http\:\/\/www\.[a-z]{6,19}\.com\/
 http\:\/\/www\.[a-z]{6,19}\.co\.uk\/

** With s replacing space and backslashes removed, spaced in groups:

 111111111   222  33333333333  4  5555555555   6  777  8  99  10
 http        ://  [a-z]{3,4}   .  [a-z]{3,4}
 http        ://  [a-z]{5,19}  .  com          /
 [a-z]{4,5}  ://  [a-z]{3}     .  [a-z]{6,19}  .  com
 http        ://  www          .  [a-z]{6,19}  .  com  /
 http        ://  www          .  [a-z]{6,19}  .  co   .  uk  /

--> :// is very special in col 2

 111111111   222  33333333333  4  5555555555   6  777  8  99  10
 http        ://  [a-z]{3,4}   .  [a-z]{3,4}
 http        ://  [a-z]{5,19}  .  com          /
 [a-z]{4,5}  ://  [a-z]{3}     .  [a-z]{6,19}  .  com
 http        ://  www          .  [a-z]{6,19}  .  com  /
 http        ://  www          .  [a-z]{6,19}  .  co   .  uk  /

--> /, when present is always last in RH group

 111111111   222  33333333333  4  5555555555   6  777  8  99  10
 http        ://  [a-z]{3,4}   .  [a-z]{3,4}
 http        ://  [a-z]{5,19}  .  com                         /
 [a-z]{4,5}  ://  [a-z]{3}     .  [a-z]{6,19}  .  com
 http        ://  www          .  [a-z]{6,19}  .  com         /
 http        ://  www          .  [a-z]{6,19}  .  co   .  uk  /



Goal:

 111111111   222  33333333333  4  5555555555   6  777  8  99  10
 http        ://  [a-z]{3,4}   .  [a-z]{3,4}
 http        ://  [a-z]{5,19}  .  com                         /
 [a-z]{4,5}  ://  [a-z]{3}     .  [a-z]{6,19}  .  com
 http        ://  www          .  [a-z]{6,19}  .  com         /
 http        ://  www          .  [a-z]{6,19}  .  co   .  uk  /

URLs 1 + 2:

 [a-z]{3,4}\.[a-z]{2,4}
 [a-z]{5,19}\.com\/
 [a-z]{3,4}[\.\/\:]{1,3}[a-z]{3,19}\.[a-z]{3,4}
 http\:\/\/[a-z]{5,19}\.com\/
 [a-z]{4,5}\:\/\/[a-z]{3}\.[a-z]{6,19}\.com
 http\:\/\/www\.[a-z]{6,19}\.com\/
 http\:\/\/www\.[a-z]{6,19}\.co\.uk\/

** With s replacing space and backslashes removed, spaced in groups:

 11111111111  2222222222  33333333333  4  55555555555  6  777  8  99  10
 [a-z]{3,4}   .           [a-z]{2,4}
 [a-z]{5,19}  .           com          /
 [a-z]{3,4}   [./:]{1,3}  [a-z]{3,19}  .  [a-z]{3,4}
 http         ://         [a-z]{5,19}  .  com          /
 [a-z]{4,5}   ://         [a-z]{3}     .  [a-z]{6,19}  .  com
 http         ://         www          .  [a-z]{6,19}  .  com  /
 http         ://         www          .  [a-z]{6,19}  .  co   .  uk  /

Goal:

 11111111111  2222222222  33333333333  4  55555555555  6  777  8  99  10
                          [a-z]{3,4}   .  [a-z]{2,4}
                          [a-z]{5,19}  .  com                         /
 [a-z]{3,4}   [./:]{1,3}  [a-z]{3,19}  .  [a-z]{3,4}
 http         ://         [a-z]{5,19}  .  com                         /
 [a-z]{4,5}   ://         [a-z]{3}     .  [a-z]{6,19}  .  com
 http         ://         www          .  [a-z]{6,19}  .  com         /
 http         ://         www          .  [a-z]{6,19}  .  co   .  uk  /



TELEPHONES 2

 \(\d{3,4}\) \d{3,4} \d{4}
 \+\d{1,2} \d{2,3} \d{3,4} \d{4}

** With s replacing space and backslashes removed, spaced in groups:

 1  2222222  3  4444444  5555555   666666  77777  8888
 (  d{3,4}   )  s        d{3,4}    s       d{4}
 +  d{1,2}   s  d{2,3}   s         d{3,4}  s      d{4}

-->  Space at pos -2

 1  2222222  3  4444444  5555555   666666  77777  8888
 (  d{3,4}   )  s        d{3,4}            s      d{4}
 +  d{1,2}   s  d{2,3}   s         d{3,4}  s      d{4}

-->  Space at pos -2 within left group

 1  2222222  3  4444444  5555555   666666  77777  8888
 (  d{3,4}   )           s         d{3,4}  s      d{4}
 +  d{1,2}   s  d{2,3}   s         d{3,4}  s      d{4}



Goal

 1  2222222  3  4444444  5555555   666666  77777  8  9999
 (  d{3,4}   )  s        d{3,4}    s       d{4}
 +  d{1,2}      s        d{2,3}    s       d{3,4} s  d{4}





TELEPHONES 5

 \d{3} \d{3} \d{4}
 \d{3}\-\d{3}\-\d{4}
 1 \d{3} \d{3} \d{4}
 \(\d{3}\) \d{3} \d{4}

** With s replacing space and backslashes removed, spaced in groups:

 1111  2222  3333  4  5555  6  7777
 d{3}  s     d{3}  s  d{4}
 d{3}  -     d{3}  -  d{4}
 1     s     d{3}  s  d{3}  s  d{4}
 (     d{3}  )     s  d{3}  s  d{4}

--> Last group is always 4 digits

 1111  2222  3333  4  5555  6  7777
 d{3}  s     d{3}  s           d{4}
 d{3}  -     d{3}  -           d{4}
 1     s     d{3}  s  d{3}  s  d{4}
 (     d{3}  )     s  d{3}  s  d{4}

--> Group -2 with left part is always 3 digits

 1111  2222  3333  4  5555  6  7777
 d{3}  s              d{3}  s  d{4}
 d{3}  -              d{3}  -  d{4}
 1     s     d{3}  s  d{3}  s  d{4}
 (     d{3}  )     s  d{3}  s  d{4}

--> Last of left of left is always space or hyphen

 1111  2222  3333  4  5555  6  7777
 d{3}              s  d{3}  s  d{4}
 d{3}              -  d{3}  -  d{4}
 1     s     d{3}  s  d{3}  s  d{4}
 (     d{3}  )     s  d{3}  s  d{4}

--> Within left three, there is always a block of 3 digits

 1  2  3333  4  5  6666  7  8888
       d{3}     s  d{3}  s  d{4}
       d{3}     -  d{3}  -  d{4}
 1  s  d{3}     s  d{3}  s  d{4}
 (     d{3}  )  s  d{3}  s  d{4}


Goal:

 1  2222  3  4444  5  6666  7  8888
             d{3}  s  d{3}  s  d{4}
             d{3}  -  d{3}  -  d{4}
    1     s  d{3}  s  d{3}  s  d{4}
 (  d{3}  )        s  d{3}  s  d{4}



TELS 1-5

 \d{3} \d{3} \d{4}
 \d{3,4}[\-\.]\d{3}[\-\.]\d{4}
 1 \d{3} \d{3} \d{4}
 \(\d{3,4}\) \d{3,4} \d{4}
 \+\d{1,2} \d{2,3} \d{3,4} \d{4}


** With s replacing space and backslashes removed, spaced in groups:

 111111  222222  3333  444444  555555  666666  7777  8888
 d{3}    s       d{3}  s       d{4}
 d{3,4}  [-.]    d{3}  [-.]    d{4}
 1       s       d{3}  s       d{3}    s       d{4}
 (       d{3,4}  )     s       d{3,4}  s       d{4}
 +       d{1,2}  s     d{2,3}  s       d{3,4}  s     d{4}


Goal:

 1  222222  3333  444444  555555  6666  777777  8  9999
    d{3}          s       d{3}    s     d{4}
    d{3,4}  [-.]          d{3}    [-.]  d{4}
    1             s       d{3}    s     d{3}    s  d{4}
 (  d{3,4}  )     s       d{3,4}  s     d{4}
 +  d{1,2}        s       d{2,3}  s     d{3,4}  s  d{4}
"""

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import os
import sys
import unittest

from collections import OrderedDict

try:
    import pandas
except:
    pandas = None
pd = pandas

#from artists.miro.writabletestcase import WritableTestCase
from tdda.rexpy.relib import re
from tdda.rexpy import *
from tdda.rexpy.rexpy import Coverage


class TestUtilityFunctions(unittest.TestCase):
    def test_signature(self):
        self.assertEqual(signature([]), '')

        self.assertEqual(signature([('c', 1)]), 'c')

        self.assertEqual(signature([('C', 3), ('.', 1),
                                    ('C', 2), ('.', 1),
                                    ('C', 2)]), 'C.C.C')

        self.assertEqual(signature([('C', 8), ('.', 1),
                                    ('C', 4), ('.', 1),
                                    ('C', 4), ('.', 1),
                                    ('C', 4), ('.', 1),
                                    ('C', 12)]), 'C.C.C.C.C')

        self.assertEqual(signature([('.', 1),
                                    ('C', 4), ('.', 1), (' ', 1),
                                    ('C', 3), (' ', 1),
                                    ('C', 4)]), '.C. C C')

        self.assertEqual(signature([('C', 4), ('.', 1),
                                    ('C', 2), ('.', 1),
                                    ('C', 5), ('.', 1),
                                    ('C', 2), ('.', 1,),
                                    ('C', 2), (' ', 1), ('.', 1),
                                    ('C', 5)]), 'C.C.C.C.C .C')

        self.assertEqual(signature([('C', 4), ('.', 1),
                                    ('C', 2), ('.', 1),
                                    ('C', 5), ('.', 1),
                                    ('C', 2), ('.', 1),
                                    ('C', 2), ('*', 1), ('.', 1),
                                    ('C', 5)]), 'C.C.C.C.C*.C')


    def test_get_omnipresent_at_pos(self):
        c = {
            ('a', 1, 1, 'fixed'): {1: 7, -1: 7, 3: 4},
            ('b', 1, 1, 'fixed'): {2: 6, 3: 4},
            ('c', 1, 1): {3: 1, 4: 2, 2: 7},
            ('d', 1, 1, 'fixed'): {1: 1}
        }

        self.assertEqual(get_omnipresent_at_pos(c, 7),
                         [((u'a', 1, 1, u'fixed'), -1),
                          ((u'a', 1, 1, u'fixed'), 1),
                          ((u'c', 1, 1), 2)])

        self.assertEqual(get_omnipresent_at_pos(c, 6),
                         [((u'b', 1, 1, u'fixed'), 2)])


        self.assertEqual(get_omnipresent_at_pos(c, 5), [])

        self.assertEqual(get_omnipresent_at_pos({}, 1), [])

        self.assertEqual(get_omnipresent_at_pos({}, 0), [])

    def test_length_stats(self):
        # Testing with strings, but works with lists etc. too
        self.assertEqual(length_stats(['abc', 'def', 'ghi']), (True, 3))
        self.assertEqual(length_stats(['a', 'def', 'gh']), (False, 3))
        self.assertEqual(length_stats([]), (True, 0))

    def test_left_parts1(self):
        # For this test, the fragments are always present in the positions
        p1 = [('a',), ('b',), ('c',), ('d',), ('e',), ('f',)]
        p2 = [('A',), ('b',), ('C',), ('d',), ('E',)]
        p3 = [('.',), ('b',), ('.',), ('d',)]
        fixed = [(('b',), 1), (('d',),  3)]
        expected = [
            [[('a',)],          [('A',)],  [('.',)]],
            [[('b',)],          [('b',)],  [('b',)]],
            [[('c',)],          [('C',)],  [('.',)]],
            [[('d',)],          [('d',)],  [('d',)]],
            [[('e',), ('f',)],  [('E',)],  []]
        ]
        self.assertEqual(left_parts([p1, p2, p3], fixed), expected)

    def test_left_parts2(self):
        # For this test, the fragments are always present in the positions
        p1 = [('a',), ('b',), ('c',), ('d',), ('e',), ('f',)]
        p2 = [('a',), ('B',), ('C',), ('d',), ('E',)]
        p3 = [('a',), ('b',), ('c',), ('d',)]
        fixed = [(('a',), 0), (('c',),  3)]
        expected = [
            [[('a',)],           [('a',)],          [('a',)]],
            [[('b',), ('c',)],   [('B',), ('C',)],  [('b',), ('c',)]],
            [[('d',)],           [('d',)],          [('d',)]],
            [[('e',), ('f',)],   [('E',)],          []]
        ]
        self.assertEqual(left_parts([p1, p2, p3], fixed), expected)

    def test_right_parts1(self):
        p1 = [('F',), ('e',), ('d',), ('c',), ('b',), ('a',)]
        p2 = [('E',), ('d',), ('C',), ('b',), ('A',)]
        p3 = [('d',), ('.',), ('b',), ('.',)]
        fixed = [(('b',), 2), (('d',),  4)]
        expected = [
            [[('F',), ('e',)],  [('E',)],  []],
            [[('d',)],          [('d',)],  [('d',)]],
            [[('c',)],          [('C',)],  [('.',)]],
            [[('b',)],          [('b',)],  [('b',)]],
            [[('a',)],          [('A',)],  [('.',)]],
        ]
        self.assertEqual(right_parts([p1, p2, p3], fixed), expected)

    def test_right_parts2(self):
        p1 = [('F',), ('e',), ('d',), ('c',), ('b',), ('a',)]
        p2 = [('E',), ('d',), ('C',), ('b',), ('a',)]
        p3 = [('d',), ('.',), ('.',), ('a',)]
        fixed = [(('a',), 1), (('d',),  4)]
        expected = [
            [[('F',), ('e',)],  [('E',)],  []],
            [[('d',)],          [('d',)],  [('d',)]],
            [[('c',), ('b',)],  [('C',), ('b',)],  [('.',), ('.',)]],
            [[('a',)],          [('a',)],  [('a',)]],
        ]
        self.assertEqual(right_parts([p1, p2, p3], fixed), expected)

    def test_ndigits(self):
        self.assertEqual(ndigits(1, 1), '1')
        self.assertEqual(ndigits(2, 1), '1' * 2)
        self.assertEqual(ndigits(1, 2), '2')
        self.assertEqual(ndigits(2, 2), '2' * 2)
        self.assertEqual(ndigits(0, 2), '')
        self.assertEqual(ndigits(11, 9), '9' * 11)
        self.assertEqual(ndigits(9, 11), 'B' * 9)
        self.assertEqual(ndigits(9, 10), 'A' * 9)
        self.assertEqual(ndigits(10,9), '9' * 10)
        self.assertEqual(ndigits(10,10), 'A' * 10)
        self.assertEqual(ndigits(5,35), 'Z' * 5)
        # Not really designed for more than 35 parts.
        # But if it happens, just carries on
        self.assertEqual(ndigits(5, 36), chr(ord('Z') + 1) * 5)
        # Also not supposed to take digits below 1.
        # but...
        self.assertEqual(ndigits(4, 0), '0' * 4)

    def test_aligned_parts1(self):
        x = Extractor([])
        part1 = [
                    [('a', 1, 1, 'fixed'), ('b', 1, 1, 'fixed')],
                    [('c', 1, 1, 'fixed')],
                ]
        part2 = [
                    [('/', 1, 1, 'fixed')],
                    [('/', 1, 1, 'fixed')],
                ]
        parts = [part1, part2]
        expected = '\n'.join(['|1 2|3|',
                              ' a b /',
                              ' c   /'])
        self.assertEqual(x.aligned_parts(parts), expected)

    def test_aligned_parts2(self):
        x = Extractor([])
        part1 = [
                    [('a', 1, 1, 'fixed'), ('b', 1, 1, 'fixed')],
                    [('c', 1, 1, 'fixed')],
                ]
        part2 = [
                    [],
                    [('/', 1, 1, 'fixed')],
                ]
        part3 = [
                    [('.', 1, 1, 'fixed')],
                    [],
                ]
        part4 = [
                    [('!', 1, 1, 'fixed')],
                    [('!', 1, 1, 'fixed')],
                ]
        parts = [part1, part2, part3, part4]
        expected = '\n'.join(['|1 2|3|4|5|',
                              ' a b   . !',
                              ' c   /   !'])
        self.assertEqual(x.aligned_parts(parts), expected)



class TestHelperMethods(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.x = Extractor([])

    def test_coarse_character_classification(self):
        x = self.x
        for i in range(ord('A'), ord('Z') + 1):
            self.assertEqual(x.coarse_classify_char(chr(i)), C)
        for i in range(ord('a'), ord('z') + 1):
            self.assertEqual(x.coarse_classify_char(chr(i)), C)
        for i in range(ord('0'), ord('9') + 1):
            self.assertEqual(x.coarse_classify_char(chr(i)), C)
        for c in ' \t\r\n\f\v':
            self.assertEqual(x.coarse_classify_char(c), ' ')
        for c in '!"#$%&' + "'" + '()*+,-.' + '\\' + ':;<=>?@[\]^_`{|}~':
            self.assertEqual(x.coarse_classify_char(c), '.')
        for i in range(0, 0x1c):  # 1C to 1F are considered whitespace
                                  # in unicode
            c = chr(i)
            if not c in '\t\r\n\f\v':
                self.assertEqual(x.coarse_classify_char(c), '*')

    def test_coarse_string_classification(self):
        x = self.x
        for i in range(ord('A'), ord('Z') + 1):
            self.assertEqual(x.coarse_classify_char(chr(i)), C)
        for i in range(ord('a'), ord('z') + 1):
            self.assertEqual(x.coarse_classify_char(chr(i)), C)
        for i in range(ord('0'), ord('9') + 1):
            self.assertEqual(x.coarse_classify_char(chr(i)), C)
        for c in ' \t\r\n\f\v':
            self.assertEqual(x.coarse_classify_char(c), ' ')
        for c in ' \t\r\n\f\v':
            self.assertEqual(x.coarse_classify_char(c), ' ')
        for c in '!"#$%&' + "'" + '()*+,-.' + '\\' + ':;<=>?@[\]^_`{|}~':
            self.assertEqual(x.coarse_classify_char(c), '.')
        for i in range(0, 0x1c):  # 1C to 1F are considered whitespace
                                  # in unicode
            c = chr(i)
            if not c in '\t\r\n\f\v':
                self.assertEqual(x.coarse_classify_char(c), '*')
        self.assertEqual(x.coarse_classify_char(chr(127)), '*')

    def test_coarse_classification(self):
        x = self.x
        self.assertEqual(x.coarse_classify('255-SI-32'),
                                         CtoUC('CCC.CC.CC'))
        guid = '1f65c9e8-cf9a-4e53-b7d0-c48a26a21b7c'
        sig  = CtoUC('CCCCCCCC.CCCC.CCCC.CCCC.CCCCCCCCCCCC')
        self.assertEqual(x.coarse_classify(guid), sig)
        self.assertEqual(x.coarse_classify('(0131) 123 4567'),
                                           CtoUC('.CCCC. CCC CCCC'))
        self.assertEqual(x.coarse_classify('2016-01-02T10:11:12 +0300z'),
                                           CtoUC('CCCC.CC.CCCCC.CC.CC .CCCCC'))
        self.assertEqual(x.coarse_classify('2016-01-02T10:11:12\a+0300z'),
                                           CtoUC('CCCC.CC.CCCCC.CC.CC*.CCCCC'))

    def test_run_length_encoding(self):
        self.assertEqual(run_length_encode(''), ())

        self.assertEqual(run_length_encode('c'), (('c', 1),))

        self.assertEqual(run_length_encode('CCC.CC.CC'),
                                           (('C', 3), ('.', 1),
                                            ('C', 2), ('.', 1),
                                            ('C', 2)))

        sig  = 'CCCCCCCC.CCCC.CCCC.CCCC.CCCCCCCCCCCC'
        self.assertEqual(run_length_encode(sig), (('C', 8), ('.', 1),
                                         ('C', 4), ('.', 1),
                                         ('C', 4), ('.', 1),
                                         ('C', 4), ('.', 1),
                                         ('C', 12)))

        self.assertEqual(run_length_encode('.CCCC. CCC CCCC'),
                         (('.', 1),
                          ('C', 4), ('.', 1), (' ', 1),
                          ('C', 3), (' ', 1),
                          ('C', 4)))

        self.assertEqual(run_length_encode('CCCC.CC.CCCCC.CC.CC .CCCCC'),
                         (('C', 4), ('.', 1),
                          ('C', 2), ('.', 1),
                          ('C', 5), ('.', 1),
                          ('C', 2), ('.', 1,),
                          ('C', 2), (' ', 1), ('.', 1),
                          ('C', 5)))

        self.assertEqual(run_length_encode('CCCC.CC.CCCCC.CC.CC*.CCCCC'),
                         (('C', 4), ('.', 1),
                          ('C', 2), ('.', 1),
                          ('C', 5), ('.', 1),
                          ('C', 2), ('.', 1),
                          ('C', 2), ('*', 1), ('.', 1),
                          ('C', 5)))

    def test_cleaning(self):
        examples = ['123-AB-321', ' 123-AB-321', '', None, '321-BA-123 ']
        keys = ['123-AB-321', '321-BA-123']
        x = Extractor(examples, remove_empties=True, strip=True)
        self.assertEqual(set(x.example_freqs.keys()), set(keys))
        self.assertEqual(x.n_stripped, 2)
        self.assertEqual(x.n_empties, 1)
        self.assertEqual(x.n_nulls, 1)
        items = x.example_freqs.keys()
        self.assertEqual(set(items), {'123-AB-321', '321-BA-123'})

    def test_cleaning_keep_empties(self):
        examples = ['123-AB-321', ' 123-AB-321', '', None, '321-BA-123 ']
        keys = ['123-AB-321', '321-BA-123', '']
        x = Extractor(examples, remove_empties=False, strip=True)
        self.assertEqual(set(x.example_freqs.keys()), set(keys))
        self.assertEqual(x.n_stripped, 2)
        self.assertEqual(x.n_empties, 0)
        self.assertEqual(x.n_nulls, 1)
        items = x.example_freqs.keys()
        self.assertEqual(set(items), {'123-AB-321', '321-BA-123', ''})

    def test_cleaning_no_strip_no_empties(self):
        examples = ['123-AB-321', ' 123-AB-321', '', None, '321-BA-123 ']
        keys = ['123-AB-321', '', ' 123-AB-321', '321-BA-123 ']
        x = Extractor(examples, remove_empties=False, strip=False)
        self.assertEqual(set(x.example_freqs.keys()), set(keys))
        self.assertEqual(x.n_stripped, 0)
        self.assertEqual(x.n_empties, 0)
        self.assertEqual(x.n_nulls, 1)
        items = x.example_freqs.keys()
        self.assertEqual(set(items), set(keys))

    def test_batch_rle_extract_single(self):
        examples = ['123-AB-321', '123-AB-321',  None, '321-BA-123']
        x = Extractor(examples)
        freqs = x.results.rle_freqs
        self.assertEqual(len(freqs), 1)
        key = ((C, 3), ('.', 1), (C, 2), ('.', 1), (C, 3))
        self.assertEqual(list(freqs.keys()), [key])
        self.assertEqual(freqs[key], 2)

    def test_batch_rle_extract_pair(self):
        examples = ['123-AB-321', '12-AB-4321', None, '321-BA-123']
        x = Extractor(examples)
        freqs = x.results.rle_freqs
        self.assertEqual(len(freqs), 2)
        key1 = ((C, 3), ('.', 1), (C, 2), ('.', 1), (C, 3))
        key2 = ((C, 2), ('.', 1), (C, 2), ('.', 1), (C, 4))
        keys = [key1, key2]
        self.assertEqual(set(freqs.keys()), set(keys))
        self.assertEqual(freqs[key1], 2)
        self.assertEqual(freqs[key2], 1)

    def test_rle2re(self):
        Cats = self.x.Cats
        rle = (('C', 3), ('.', 1), ('C', 2), ('.', 1), ('C', 3))
        an = Cats.AlphaNumeric.re_string
        punc = Cats.Punctuation.re_string
        regex = self.x.rle2re(rle)
        self.assertEqual(regex,
                         '^%s{3}%s%s{2}%s%s{3}$' % (an, punc, an, punc, an))
        cre = re.compile(regex)
        for s in ['123-AB-321', '321-BA-123']:
            self.assertIsNotNone(re.match(cre, s))

    def test_vrle(self):
        key1 = (('C', 3), ('.', 1), ('C', 2), ('.', 1), ('C', 3))
        key2 = (('C', 2), ('.', 1), ('C', 2), ('.', 1), ('C', 4))
        keys = [key1, key2]
        range_rles = to_vrles(keys)
        self.assertEqual(range_rles, [(('C', 2, 3), ('.', 1, 1),
                                       ('C', 2, 2), ('.', 1, 1),
                                       ('C', 3, 4))])

    def test_vrle2re(self):
        key1 = (('C', 3), ('.', 1), ('C', 2), ('.', 1), ('C', 3))
        key2 = (('C', 2), ('.', 1), ('C', 2), ('.', 1), ('C', 4))
        keys = [key1, key2]
        rrle = [('C', 2, 3), ('.', 1, 1),
                ('C', 2, 2), ('.', 1, 1),
                ('C', 3, 4)]
        Cats = Extractor([]).Cats
        an = Cats.AlphaNumeric.re_string
        punc = Cats.Punctuation.re_string
        expected = '^%s{2,3}%s%s{2}%s%s{3,4}$' % (an, punc, an, punc, an)
        x = Extractor([])
        self.assertEqual(x.vrle2re(rrle), expected)

    def test_sort_by_len(self):
        # Really designed for sorting lists/tuples, but everything with
        # a length will behave the same
        x = Extractor(['a'])
        self.assertEqual(x.sort_by_length([]), [])
        self.assertEqual(x.sort_by_length(['a']), ['a'])
        self.assertEqual(x.sort_by_length(['a', 'ab']), ['a', 'ab'])
        self.assertIn(x.sort_by_length(
                ['abcdef', '', 'aa', 'bb', 'ccc', 'zzyy', '1']),
                (['', '1', 'aa', 'bb', 'ccc', 'zzyy', 'abcdef'],
                 ['', '1', 'bb', 'aa', 'ccc', 'zzyy', 'abcdef']))

    def test_split_left(self):
        x = Extractor(['a'])
        http = [('http', 1, 1, 'fixed'),
                ('://', 1, 1, 'fixed'),
                ('a', 1, 5),
                ('.', 1, 1, 'fixed'),
                ('com', 1, 1, 'fixed')]
        https = [('https', 1, 1, 'fixed'),
                 ('://', 1, 1, 'fixed'),
                 ('a', 1, 4),
                 ('#', 1, 1, 'fixed'),
                 ('co', 1, 1, 'fixed'),
                 ('.', 1, 1, 'fixed'),
                 ('uk', 1, 1, 'fixed'),
                 ('/', 1, 1, 'fixed')]
        patterns = [http, https]
        results = x.merge_fixed_omnipresent_at_pos(patterns)
        expected = [
                [
                    [(u'http', 1, 1, u'fixed')],
                    [(u'https', 1, 1, u'fixed')]
                ],

                [
                    [(u'://', 1, 1, u'fixed')],
                    [(u'://', 1, 1, u'fixed')],
                ],

                [
                    [(u'a', 1, 5),
                     (u'.', 1, 1, u'fixed'),
                     (u'com', 1, 1, u'fixed')],

                    [(u'a', 1, 4),
                     (u'#', 1, 1, u'fixed'),
                     (u'co', 1, 1, u'fixed'),
                     (u'.', 1, 1, u'fixed'),
                     (u'uk', 1, 1, u'fixed'),
                     (u'/', 1, 1, u'fixed')]
                ],
        ]
        self.assertEqual(results, expected)


class TestExtraction(unittest.TestCase):

    def check_result(self, result, rexes, examples):
        self.assertEqual(result, rexes)
        compiled = [re.compile(r) for r in rexes]
        self.assertEqual(all(any(re.match(rex, x) for rex in compiled)
                             for x in examples if x),
                         True)

    def test_re_pqs_id(self):
        # Space in second string should cause \s* at start, no?
        iids = ['123-AB-321', ' 12-AB-4321', '', None, '321-BA-123 ']
        rexes = [r'^\s*\d{2,3}\-[A-Z]{2}\-\d{3,4}\s*$']
        x = extract(iids, strip=True, remove_empties=True)
        self.check_result(x, rexes, iids)

    def test_re_pqs_id_with_dash(self):
        iids = ['123-AB-321', '12-AB-4321', '', None, '321-BA-123']
        rexes = [r'^$', r'^\d{2,3}[A-Z\-][A-Z]{2}[A-Z\-]\d{3,4}$']
        x = extract(iids, extra_letters='-')
        self.check_result(x, rexes, iids)
        # Test result has changed with improved behaviour.
        # Now us spots that the base sequence is the same
        # Previously this found this:
        #        [r'^\s*[A-Z0-9\-]{10}\s*$'])

    def test_re_pqs_id_with_dash2(self):
        iids = ['123-AB-321', 'AB-1B-4A21', None, '321-BA-1A23']
        rexes = [r'^[A-Z0-9\-]{10,11}$']
        x = extract(iids, extra_letters='-')
        self.check_result(x, rexes, iids)

    def test_re_pqs_id_with_underscore(self):
        iids = ['123_AB_321', 'AB_1B_4A21', None, '321_BA_1A23ab2rj']
        rexes = [r'^[A-Za-z0-9\_]+$']
        x = extract(iids, extra_letters='_')
        self.check_result(x, rexes, iids)

    def test_re_pqs_id_with_underscores2(self):
        iids = ['123_AB_321', 'AB_1B_4A21', None, '321_BA_1A23ab2rj']
        rexes = [r'^[A-Za-z0-9\_]+$']
        x = extract(iids, extra_letters='_-')
        self.check_result(x, rexes, iids)

    def test_re_pqs_id_with_underscores3(self):
        iids = ['123-AB_321', 'AB_1B-4A21', None, '321_BA_1A23ab2rj']
        rexes = [r'^[A-Za-z0-9\_\-]+$']
        x = extract(iids, extra_letters='_-.')
        self.check_result(x, rexes, iids)

    def test_re_uuid(self):
        uuids = ['1f65c9e8-cf9a-4e53-b7d0-c48a26a21b7c',
                 '88888888-4444-4444-4444-cccccccccccc',
                 'aaaaaaaa-1111-0000-9999-0123456789ab',
                 '22ecc913-68eb-4cfb-a9cc-18df44137a4c',
                 '634962c3-8bc1-4b51-b36d-3dcbfdc92b63',
                 'aac65b99-92ff-11e6-b97d-b8f6b118f191',
                 '6fa459ea-ee8a-3ca4-8f4e-db77e160355e',
                 '886313e1-3b8a-e372-9b90-0c9aee199e5d']
        rexes = ['^[0-9a-f]{8}\\-[0-9a-f]{4}\\-[0-9a-f]{4}\\-'
                 '[0-9a-f]{4}\\-[0-9a-f]{12}$']
        x = extract(uuids)
        self.check_result(x, rexes, uuids)

    def test_re_uuid4(self):
        uuid4s = ['2ffb8eaa-dd75-41c2-aca6-0444914b8713',
                  'f69c5651-0b97-4909-b896-8ef0891f81ff',
                  '13f984fe-65db-4646-99d4-a93c06f78472',
                  '50a886d2-78c9-4b7b-81a9-25caf4deb212',
                  '2d4429df-9a80-b581-9565-27880ce171b0',
                  '857e0ec6-1511-478b-93a3-15ac9212fd0d']

        rexes = [r'^[0-9a-f]{8}\-[0-9a-f]{4}\-[0-9a-f]{4}\-'
                 r'[0-9a-f]{4}\-[0-9a-f]{12}$']

        x = extract(uuid4s, variableLengthFrags=True)       # Not refining down yet
        self.check_result(x, rexes, uuid4s)

    tels1 = [
        '(0131) 496 0091',
        '(0131) 496 0161',
        '(0131) 496 0107',
        '(020) 7946 0106',
        '(020) 7946 0642',
        '(0141) 496 0927',
        '(0141) 496 0595',
        '(0141) 496 0236',
        '(0141) 496 0324'
    ]
    def test_tels1(self):
        x = extract(self.tels1)
        self.check_result(x, [r'^\(\d{3,4}\) \d{3,4} \d{4}$'], self.tels1)


    tels2 = [
        '+44 131 496 0091',
        '(0131) 496 0161',
        '+44 131 496 0107',
        '(020) 7946 0106',
        '+44 20 7946 0642',
        '+44 141 496 0927',
        '(0141) 496 0595',
        '+1 202 555 0181',
        '(0141) 496 0324',
    ]
    def test_tels2(self):
        x = set(extract(self.tels2))
        rexes = set([r'^\+\d{1,2} \d{2,3} \d{3,4} \d{4}$',
                     r'^\(\d{3,4}\) \d{3,4} \d{4}$'])
        self.check_result(x, rexes, self.tels2)

    def test_coverage_tels2(self):
        x = Extractor(self.tels2)
        rex = x.results.rex
        d = dict(zip(rex, x.coverage()))
        self.assertEqual(d,
                         {
                            r'^\+\d{1,2} \d{2,3} \d{3,4} \d{4}$': 5,
                            r'^\(\d{3,4}\) \d{3,4} \d{4}$': 4,
                         })
        expected = set([r'^\+\d{1,2} \d{2,3} \d{3,4} \d{4}$',
                        r'^\(\d{3,4}\) \d{3,4} \d{4}$'])
        self.check_result(set(rex), expected, self.tels2)

    def test_incremental_coverage_tels2(self):
        x = Extractor(self.tels2)
        rex = x.results.rex
        od = x.incremental_coverage()
        self.assertEqual(od,
                         OrderedDict((
                            (r'^\+\d{1,2} \d{2,3} \d{3,4} \d{4}$', 5),
                            (r'^\(\d{3,4}\) \d{3,4} \d{4}$', 4),
                         )))

    def test_coverage_tels2_dedup(self):
        x = Extractor(self.tels2 + self.tels2[:6])
        rex = x.results.rex
        self.assertEqual(type(rex), list)
        d = dict(zip(rex, x.coverage()))
        self.assertEqual(d,
                         {
                            r'^\+\d{1,2} \d{2,3} \d{3,4} \d{4}$': 9,
                            r'^\(\d{3,4}\) \d{3,4} \d{4}$': 6,
                         })
        d = dict(zip(rex, x.coverage(dedup=True)))
        self.assertEqual(d,
                         {
                            r'^\+\d{1,2} \d{2,3} \d{3,4} \d{4}$': 5,
                            r'^\(\d{3,4}\) \d{3,4} \d{4}$': 4,
                         })

    def test_incremental_coverage_tels2_dedup(self):
        x = Extractor(self.tels2 + self.tels2[:6])
        rex = x.results.rex
        od = x.incremental_coverage(dedup=True)
        self.assertEqual(od,
                         OrderedDict((
                            (r'^\+\d{1,2} \d{2,3} \d{3,4} \d{4}$', 5),
                            (r'^\(\d{3,4}\) \d{3,4} \d{4}$', 4),
                         )))

    tels3 = [
        '0131-222-9876',
        '0207.987.2287'
    ]

    def test_tels3(self):
        r = extract(self.tels3, as_object=True)
        self.assertEqual(r.warnings, [])
        self.assertEqual(r.results.refrags,
                         [[(u'\\d{4}', True),
                           (u'[\\-\\.]', False),
                           (u'\\d{3}', True),
                           (u'[\\-\\.]', False),
                            (u'\\d{4}', True),
                          ]])
        self.check_result(r.results.rex, [r'^\d{4}[\-\.]\d{3}[\-\.]\d{4}$'],
                          self.tels3)

    tels4 = [
        '222-678-8834',
        '321-123-1234',
        '212-988-0321',
        '191-777-2043'
    ]
    def test_tels4(self):
        x = extract(self.tels4)
        self.check_result(x, [r'^\d{3}\-\d{3}\-\d{4}$'], self.tels4)


    tels5 = [
        '212-988-0321',
        '987-654-3210',
        '476 123 8829',
        '123 456 7890',
        '1 701 734 9288',
        '1 177 441 7712',
        '(617) 222 0529',
        '(222) 111 9276',
    ]
    def test_tels5(self):
#        print(Extractor(self.tels5))
        x = extract(self.tels5)
        rexes = [
            r'^\d{3} \d{3} \d{4}$',
            r'^\d{3}\-\d{3}\-\d{4}$',
            r'^\(\d{3}\) \d{3} \d{4}$',
            r'^1 \d{3} \d{3} \d{4}$',
        ]
        self.check_result(x, rexes, self.tels5)

    def test_tels_1to5(self):
        tels = self.tels1 + self.tels2 + self.tels3 + self.tels4 + self.tels5
        x = extract(tels)
        rexes = [
            r'^\d{3} \d{3} \d{4}$',
            r'^\d{3,4}[\-\.]\d{3}[\-\.]\d{4}$',
            r'^\(\d{3,4}\) \d{3,4} \d{4}$',
            r'^1 \d{3} \d{3} \d{4}$',
            r'^\+\d{1,2} \d{2,3} \d{3,4} \d{4}$',
        ]
        self.check_result(x, rexes, tels)

    urls1 = [
        'http://www.stochasticsolutions.com/',
        'http://apple.com/',
        'http://stochasticsolutions.com/',
        'http://www.stochasticsolutions.co.uk/',
        'http://www.google.co.uk/',
        'http://www.google.com',
        'http://www.google.com/',
        'http://www.stochasticsolutions.com',
        'http://web.stochasticsolutions.com',
        'https://www.stochasticsolutions.com',
        'http://tdda.info',
        'http://web.web',
    ]

    POSTCODES = [
        'EH3 1LH',
        'G4 2BP',
    ]
    def test_POSTCODES(self):
        self.assertEqual(extract(self.POSTCODES),
                         [r'^[A-Z]{1,2}\d \d[A-Z]{2}$'])
        # Improved answer
        # Previously
        #    [r'^[A-Z0-9]{2,3} [A-Z0-9]{3}$'])


    Postcodes = [
        'EH3 1LH',
        'g4 2bp',
    ]
    def test_Postcodes(self):
        self.assertEqual(extract(self.Postcodes),
                         [r'^[A-Za-z0-9]{2,3} [A-Za-z0-9]{3}$'])

    postcodes = [
        'eh3 1lh',
        'g4 2bp',
    ]
    def test_Postcodes2(self):
        self.assertEqual(extract(self.postcodes),
                         [r'^[a-z]{1,2}\d \d[a-z]{2}$'])
        # Improved answer.
        # Previously:
        #     [r'^[a-z0-9]{2,3} [a-z0-9]{3}$'])

    postCODES = [
        'eh3 1LH',
        'g4 2BP',
    ]
    def test_postCODES(self):
        self.assertEqual(extract(self.postCODES),
                         [r'^[a-z]{1,2}\d \d[A-Z]{2}$'])
        # Improved answer.
        # Previously:
        #     [r'^[a-z0-9]{2,3} [A-Z0-9]{3}$'])

    POSTcodes = [
        'EH3 1lh',
        'G4 2bp',
    ]
    def test_POSTcodes(self):
        self.assertEqual(extract(self.POSTcodes),
                         [r'^[A-Z]{1,2}\d \d[a-z]{2}$'])
#                         [r'^[A-Z0-9]{2,3} [a-z0-9]{3}$'])


    names1 = [
        'Euclid',
        'Homer',
        'Plato',
    ]

    names1L = names1 + ['Socrates']

    names2 = [
        'Albert Einstein',
        'Richard Feynman',
        'Miles Davis',
    ]
    names3 = [
        'James Clerk Maxwell',
        'Ralph Waldo Emerson',
        'Carl Friedrich Gauss',
    ]

    namesAU = names2 + [
        'Emmy Nöther'
    ]

    namesUA = names2 + [
        'Stanisław Mazur'
    ]

    namesUAU = names3 + [
        'Émilie du Châtelet'
    ]

    names_dot_initial = [
        'John F. Kennedy',
        'George W. Bush',
    ]

    def test_names1(self):
        self.assertEqual(extract(self.names1), [r'^[A-Z][a-z]{4,5}$'])


    def test_names1L(self):
        self.assertEqual(extract(self.names1L), [r'^[A-Z][a-z]+$'])


    def test_names2(self):
        self.assertEqual(extract(self.names2),
                         [r'^[A-Z][a-z]{4,6} [A-Z][a-z]+$'])


    def test_names3(self):
        self.assertEqual(extract(self.names3),
                         [r'^[A-Z][a-z]{3,4} [A-Z][a-z]+ [A-Z][a-z]{4,6}$'])

    @unittest.skipIf(not UNICHRS, 'Unicode handling off')
    def test_namesAU(self):
        self.assertEqual(extract(self.namesAU),
                        [r'^[A-Z][a-z]+ [^\W0-9_]+$'])

    @unittest.skipIf(not UNICHRS, 'Unicode handling off')
    def test_namesAU(self):
        self.assertEqual(extract(self.namesUA),
                         [r'^[^\W0-9_]+ [A-Z][a-z]+$'])

    @unittest.skipIf(not UNICHRS, 'Unicode handling off')
    def test_namesUAU(self):
        self.assertEqual(extract(self.namesUAU),
                         [r'^[^\W0-9_]{4,6} [A-Za-z]+ [^\W0-9_]+$'])

    def test_names_dot_initial(self):
        self.assertEqual(extract(self.names_dot_initial),
                         [r'^[A-Z][a-z]{3,5} [A-Z]\. [A-Z][a-z]+$'])

    def test_names1L23(self):
        self.assertEqual(extract(self.names1L + self.names2 + self.names3),
                         [r'^[A-Z][a-z]+$',
                          r'^[A-Z][a-z]{4,6} [A-Z][a-z]+$',
                          r'^[A-Z][a-z]{3,4} [A-Z][a-z]+ [A-Z][a-z]{4,6}$'])

    def test_names1L23_dot_initial(self):
        self.assertEqual(extract(self.names1L + self.names2 + self.names3
                                 + self.names_dot_initial),
                         [r'^[A-Z][a-z]+$',
                          r'^[A-Z][a-z]{4,6} [A-Z][a-z]+$',
                          r'^[A-Z][a-z]{3,4} [A-Z][a-z]+ [A-Z][a-z]{4,6}$',
                          r'^[A-Z][a-z]{3,5} [A-Z]\. [A-Z][a-z]+$'])

    abplus = ['ab', 'abb', 'abbb', 'abbbbb', 'abbbbbbb',
              'ab', 'abbbbbbb', 'abbbb', 'ab', 'abbbb',]
    aplusbplus = ['ab', 'abb', 'abbb', 'abbbbb', 'aaaaaaaaaaabbbbbbb',
                  'aab', 'aaabb', 'aaabbbb', 'ab', 'abbbb',]

    def test_abplus(self):
        self.assertEqual(extract(self.abplus), [r'^ab+$'])

    def test_urls1(self):
        self.assertEqual(extract(self.aplusbplus), [r'^a+b+$'])

#        print()
#        x = Extractor(self.urls1, verbose=True)
#        x = Extractor(self.urls1, verbose=True)
#        print(str(x))
        self.assertEqual(set(extract(self.urls1, variableLengthFrags=False)),
                         set([r'^http\:\/\/www\.[a-z]+\.com\/$',
                              r'^http\:\/\/[a-z]{3,4}\.[a-z]{3,4}$',
                              r'^http\:\/\/[a-z]+\.com\/$',
                              r'^[a-z]{4,5}\:\/\/[a-z]{3}\.[a-z]+\.com$',
                              r'^http\:\/\/www\.[a-z]+\.co\.uk\/$',]))

        self.assertEqual(set(extract(self.urls1, variableLengthFrags=True)),
                         set([r'^http\:\/\/[a-z]{3,4}\.[a-z]{3,4}$',
                              r'^http\:\/\/www\.[a-z]+\.co\.uk\/$',
                              r'^http\:\/\/www\.[a-z]+\.com\/$',
                              r'^http\:\/\/[a-z]+\.com\/$',
                              r'^https?\:\/\/w{1,3}e?b?\.[a-z]+\.com$',]))


        # Sorted by length, these forms are:These are All:
        #
        #    C+.+C+.C+
        #    C+.+C+.C+.
        #    C+.+C+.C+.C+
        #    C+.+C+.C+.C+.
        #    C+.+C+.C+.C+.C+.

        # So: C+.+C+.C+(.C+)*.?
        # But that's a bit general!


        # Really want:
        #    '^https?\:\/\/([a-z]+\.)+[a-z]+\/?$'
        # Quite a way to go!
        #   - Small categoricals (http|https)
        #   - Related small categoricals (http|https) --> https?)
        #   - {6-19} --> +  (but when?)
        #   -  (C|C.C|C.C.C) --> (C.)+C   (i.e. concept of separation
        #                                  as opposed to terminators)


    urls2 = [
        'stochasticsolutions.com/',
        'apple.com/',
        'stochasticsolutions.com/',  # actual duplicate
        'http://www.stochasticsolutions.co.uk/',
        'http://www.google.co.uk/',
        'http://www.google.com',
        'http://www.google.com/',
        'http://www.guardian.co.uk/',
        'http://www.guardian.com',
        'http://www.guardian.com/',
        'http://www.stochasticsolutions.com',
        'web.stochasticsolutions.com',
        'https://www.stochasticsolutions.com',
        'tdda.info',
        'gov.uk',
        'http://web.web',
    ]

    def test_urls2(self):

#        print()
#        x = Extractor(self.urls2, verbose=True)
#        x = Extractor(self.urls2, variableLengthFrags=True)
#        print(str(x))
        self.assertEqual(set(extract(self.urls2, variableLengthFrags=False)), {
            r'^[a-z]{3,4}\.[a-z]{2,4}$',
            r'^[a-z]+\.com\/$',
            r'^[a-z]{3,4}[\.\/\:]{1,3}[a-z]+\.[a-z]{3}$',
            r'^[a-z]{4,5}\:\/\/www\.[a-z]+\.com$',
            r'^http\:\/\/www\.[a-z]{6,8}\.com\/$',
            r'^http\:\/\/www\.[a-z]+\.co\.uk\/$',
        })
        self.assertEqual(set(extract(self.urls2, variableLengthFrags=True)), {
            r'^[a-z]{3,4}\.[a-z]{2,4}$',
            r'^[a-z]+\.com\/$',
            r'^[a-z]{3,4}[\.\/\:]{1,3}[a-z]+\.[a-z]{3}$',
            r'^http\:\/\/www\.[a-z]{6,8}\.com\/$',
            r'^http\:\/\/www\.[a-z]+\.co\.uk\/$',
            r'^https?\:\/\/www\.[a-z]+\.com$',
        })

    def test_incremental_coverage_urls2(self):
        x = Extractor(self.urls2, variableLengthFrags=False)
        od = x.incremental_coverage()
        expected = OrderedDict((
                      (u'^[a-z]{4,5}\\:\\/\\/www\\.[a-z]+\\.com$', 4),
                      (u'^[a-z]+\\.com\\/$', 3),
                      (u'^http\\:\\/\\/www\\.[a-z]+\\.co\\.uk\\/$', 3),
                      (u'^[a-z]{3,4}[\\.\\/\\:]{1,3}[a-z]+\\.[a-z]{3}$', 2),
                      (u'^[a-z]{3,4}\\.[a-z]{2,4}$', 2),
                      (u'^http\\:\\/\\/www\\.[a-z]{6,8}\\.com\\/$', 2)))
        self.assertEqual(od, expected)
        self.assertEqual(x.n_examples(), 16)

        expected_dd = OrderedDict((
                         (u'^[a-z]{4,5}\\:\\/\\/www\\.[a-z]+\\.com$', 4),
                         (u'^http\\:\\/\\/www\\.[a-z]+\\.co\\.uk\\/$', 3),
                         (u'^[a-z]+\\.com\\/$', 2),
                         (u'^[a-z]{3,4}[\\.\\/\\:]{1,3}[a-z]+\\.[a-z]{3}$', 2),
                         (u'^[a-z]{3,4}\\.[a-z]{2,4}$', 2),
                         (u'^http\\:\\/\\/www\\.[a-z]{6,8}\\.com\\/$', 2)))
        od = x.incremental_coverage(dedup=True)
        self.assertEqual(od, expected_dd)
        self.assertEqual(x.n_examples(dedup=True), 15)


        x = Extractor(self.urls2 * 2, variableLengthFrags=False)
        doubled = OrderedDict([(k, n * 2) for k, n in expected.items()])
        od = x.incremental_coverage()
        self.assertEqual(od, doubled)
        self.assertEqual(x.n_examples(), 32)

        od = x.incremental_coverage(dedup=True)
        self.assertEqual(od, expected_dd)
        self.assertEqual(x.n_examples(dedup=True), 15)

    def test_full_incremental_coverage(self):
        freqs = {
            'One': 1,
            'one': 1,
            'two': 2,
            'three': 3,
            'four': 4,
            'five': 5,
            'six': 6,
            'seven': 7,
        }
        patterns = ['^f.+$', '^.{3}$', '^[st].+$']
        results = rexpy.rex_full_incremental_coverage(patterns, freqs)

#        for k in results:
#            print("            (u'%s',\n             rexpy.%s),"
#                  % (k, results[k]))
        expected = OrderedDict((
            (u'^[st].+$',
             Coverage(n=18,      # 2 + 3 + 6 + 7,
                            n_uniq=4,  # two, three, six, seven
                            incr=18,
                            incr_uniq=4,
                            index=2)),
            (u'^f.+$',
             Coverage(n=9,        # 4 + 5
                            n_uniq=2,   # four, five
                            incr=9,     # four (4), five (5)
                            incr_uniq=2,
                            index=0)),
            (u'^.{3}$',
             Coverage(n=10,      # 1 + 1 + 2 + 6
                            n_uniq=4,  # One, one, two, six
                            incr=2,    # One (1), one (1)
                            incr_uniq=2,
                            index=1)),
        ))
        self.assertEqual(results, expected)

        results = rexpy.rex_full_incremental_coverage(patterns, freqs,
                                                      dedup=True)
        expected = OrderedDict((
            (u'^.{3}$',
             Coverage(n=10,       # 1 + 1 + 2 + 6
                            n_uniq=4,   # One, one, two, six
                            incr=10,    # One (1), one (1), two (2), six (6)
                            incr_uniq=4,
                            index=1)),
            (u'^[st].+$',
             Coverage(n=18,       # 2 + 3 + 6 + 7,
                            n_uniq=4,   # two, three, six, seven
                            incr=10,    # three (3), seven (7)
                            incr_uniq=2,
                            index=2)),
            (u'^f.+$',
             Coverage(n=9,        # 4 + 5
                            n_uniq=2,   # four, five
                            incr=9,     # four (4), five (5)
                            incr_uniq=2,
                            index=0)),
        ))
        self.assertEqual(results, expected)

    def test_full_incremental_coverage_urls2(self):
        x = Extractor(self.urls2, variableLengthFrags=False)
        od = x.full_incremental_coverage()
        expected = OrderedDict((
            (u'^[a-z]{4,5}\:\/\/www\.[a-z]+\.com$',
             Coverage(n=4, n_uniq=4, incr=4, incr_uniq=4, index=3)),
            (u'^[a-z]+\.com\/$',
             Coverage(n=3, n_uniq=2, incr=3, incr_uniq=2, index=1)),
            (u'^http\:\/\/www\.[a-z]+\.co\.uk\/$',
             Coverage(n=3, n_uniq=3, incr=3, incr_uniq=3, index=5)),
            (u'^[a-z]{3,4}[\.\/\:]{1,3}[a-z]+\.[a-z]{3}$',
             Coverage(n=2, n_uniq=2, incr=2, incr_uniq=2, index=2)),
            (u'^[a-z]{3,4}\.[a-z]{2,4}$',
             Coverage(n=2, n_uniq=2, incr=2, incr_uniq=2, index=0)),
            (u'^http\:\/\/www\.[a-z]{6,8}\.com\/$',
             Coverage(n=2, n_uniq=2, incr=2, incr_uniq=2, index=4)),
        ))

        self.assertEqual(od, expected)
        self.assertEqual(x.n_examples(), 16)
        expected_dd = OrderedDict((
            (u'^[a-z]{4,5}\:\/\/www\.[a-z]+\.com$',
             Coverage(n=4, n_uniq=4, incr=4, incr_uniq=4, index=3)),
            (u'^http\:\/\/www\.[a-z]+\.co\.uk\/$',
             Coverage(n=3, n_uniq=3, incr=3, incr_uniq=3, index=5)),
            (u'^[a-z]+\.com\/$',
             Coverage(n=3, n_uniq=2, incr=3, incr_uniq=2, index=1)),
            (u'^[a-z]{3,4}[\.\/\:]{1,3}[a-z]+\.[a-z]{3}$',
             Coverage(n=2, n_uniq=2, incr=2, incr_uniq=2, index=2)),
            (u'^[a-z]{3,4}\.[a-z]{2,4}$',
             Coverage(n=2, n_uniq=2, incr=2, incr_uniq=2, index=0)),
            (u'^http\:\/\/www\.[a-z]{6,8}\.com\/$',
             Coverage(n=2, n_uniq=2, incr=2, incr_uniq=2, index=4)),
        ))

        od = x.full_incremental_coverage(dedup=True)
        self.assertEqual(od, expected_dd)
        self.assertEqual(x.n_examples(dedup=True), 15)

        x = Extractor(self.urls2 * 2, variableLengthFrags=False)
        doubled = OrderedDict((
            (u'^[a-z]{4,5}\:\/\/www\.[a-z]+\.com$',
             Coverage(n=8, n_uniq=4, incr=8, incr_uniq=4, index=3)),
            (u'^[a-z]+\.com\/$',
             Coverage(n=6, n_uniq=2, incr=6, incr_uniq=2, index=1)),
            (u'^http\:\/\/www\.[a-z]+\.co\.uk\/$',
             Coverage(n=6, n_uniq=3, incr=6, incr_uniq=3, index=5)),
            (u'^[a-z]{3,4}[\.\/\:]{1,3}[a-z]+\.[a-z]{3}$',
             Coverage(n=4, n_uniq=2, incr=4, incr_uniq=2, index=2)),
            (u'^[a-z]{3,4}\.[a-z]{2,4}$',
             Coverage(n=4, n_uniq=2, incr=4, incr_uniq=2, index=0)),
            (u'^http\:\/\/www\.[a-z]{6,8}\.com\/$',
             Coverage(n=4, n_uniq=2, incr=4, incr_uniq=2, index=4)),
        ))
        od = x.full_incremental_coverage()
        self.assertEqual(od, doubled)
        self.assertEqual(x.n_examples(), 32)

        od = x.full_incremental_coverage(dedup=True)

        expected_doubled_dd = OrderedDict((
            (u'^[a-z]{4,5}\:\/\/www\.[a-z]+\.com$',
             Coverage(n=8, n_uniq=4, incr=8, incr_uniq=4, index=3)),
            (u'^http\:\/\/www\.[a-z]+\.co\.uk\/$',
             Coverage(n=6, n_uniq=3, incr=6, incr_uniq=3, index=5)),
            (u'^[a-z]+\.com\/$',
             Coverage(n=6, n_uniq=2, incr=6, incr_uniq=2, index=1)),
            (u'^[a-z]{3,4}[\.\/\:]{1,3}[a-z]+\.[a-z]{3}$',
             Coverage(n=4, n_uniq=2, incr=4, incr_uniq=2, index=2)),
            (u'^[a-z]{3,4}\.[a-z]{2,4}$',
             Coverage(n=4, n_uniq=2, incr=4, incr_uniq=2, index=0)),
            (u'^http\:\/\/www\.[a-z]{6,8}\.com\/$',
             Coverage(n=4, n_uniq=2, incr=4, incr_uniq=2, index=4)),
        ))
        self.assertEqual(od, expected_doubled_dd)
        self.assertEqual(x.n_examples(dedup=True), 15)

    def test_full_incremental_coverage_urls2_var(self):
        x = Extractor(self.urls2, variableLengthFrags=True)
        od = x.full_incremental_coverage()
        expected = OrderedDict((
            (u'^https?\:\/\/www\.[a-z]+\.com$',
             Coverage(n=4, n_uniq=4, incr=4, incr_uniq=4, index=4)),
            (u'^[a-z]+\.com\/$',
             Coverage(n=3, n_uniq=2, incr=3, incr_uniq=2, index=1)),
            (u'^http\:\/\/www\.[a-z]+\.co\.uk\/$',
             Coverage(n=3, n_uniq=3, incr=3, incr_uniq=3, index=5)),
            (u'^[a-z]{3,4}[\.\/\:]{1,3}[a-z]+\.[a-z]{3}$',
             Coverage(n=2, n_uniq=2, incr=2, incr_uniq=2, index=2)),
            (u'^[a-z]{3,4}\.[a-z]{2,4}$',
             Coverage(n=2, n_uniq=2, incr=2, incr_uniq=2, index=0)),
            (u'^http\:\/\/www\.[a-z]{6,8}\.com\/$',
             Coverage(n=2, n_uniq=2, incr=2, incr_uniq=2, index=3))
        ))

        x = Extractor(self.urls2 * 2, variableLengthFrags=True)
        doubled = OrderedDict((
            (u'^https?\:\/\/www\.[a-z]+\.com$',
             Coverage(n=8, n_uniq=4, incr=8, incr_uniq=4, index=5)),
            (u'^[a-z]+\.com\/$',
             Coverage(n=6, n_uniq=2, incr=6, incr_uniq=2, index=1)),
            (u'^http\:\/\/www\.[a-z]+\.co\.uk\/$',
             Coverage(n=6, n_uniq=3, incr=6, incr_uniq=3, index=4)),
            (u'^[a-z]{3,4}[\.\/\:]{1,3}[a-z]+\.[a-z]{3}$',
             Coverage(n=4, n_uniq=2, incr=4, incr_uniq=2, index=2)),
            (u'^[a-z]{3,4}\.[a-z]{2,4}$',
             Coverage(n=4, n_uniq=2, incr=4, incr_uniq=2, index=0)),
            (u'^http\:\/\/www\.[a-z]{6,8}\.com\/$',
             Coverage(n=4, n_uniq=2, incr=4, incr_uniq=2, index=3)),
        ))
        od = x.full_incremental_coverage()
        self.assertEqual(od, doubled)
        self.assertEqual(x.n_examples(), 32)


    def test_urls2_grouped(self):

#        print()
#        x = Extractor(self.urls2, verbose=True)
#        x = Extractor(self.urls2)
#        print(str(x))
        self.assertEqual(set(extract(self.urls2, tag=True,
                                     variableLengthFrags=False)), {
            r'^([a-z]{4,5})\:\/\/www\.([a-z]+)\.com$',
            r'^http\:\/\/www\.([a-z]{6,8})\.com\/$',
            r'^([a-z]{3,4})\.([a-z]{2,4})$',
            r'^([a-z]{3,4})[\.\/\:]{1,3}([a-z]+)\.([a-z]{3})$',
            r'^([a-z]+)\.com\/$',
            r'^http\:\/\/www\.([a-z]+)\.co\.uk\/$',
        })
        self.assertEqual(set(extract(self.urls2, tag=True,
                                     variableLengthFrags=True)), {
            r'^http\:\/\/www\.([a-z]{6,8})\.com\/$',
            r'^([a-z]{3,4})\.([a-z]{2,4})$',
            r'^([a-z]{3,4})[\.\/\:]{1,3}([a-z]+)\.([a-z]{3})$',
            r'^https?\:\/\/www\.([a-z]+)\.com$',
            r'^([a-z]+)\.com\/$',
            r'^http\:\/\/www\.([a-z]+)\.co\.uk\/$',
        })

    def test_urls_all(self):

#        print()
#        x = Extractor(self.urls2, verbose=True)
#        x = Extractor(self.urls1 + self.urls2)
#        print(str(x))
        self.assertEqual(set(extract(self.urls1 + self.urls2,
                                     variableLengthFrags=False)), {
            r'^[a-z]{3,4}\.[a-z]{2,4}$',
            r'^[a-z]+\.com\/$',
            r'^[a-z]{3,4}[\.\/\:]{1,3}[a-z]+\.[a-z]{3,4}$',
            r'^http\:\/\/[a-z]+\.com\/$',
            r'^[a-z]{4,5}\:\/\/[a-z]{3}\.[a-z]+\.com$',
            r'^http\:\/\/www\.[a-z]+\.com\/$',
            r'^http\:\/\/www\.[a-z]+\.co\.uk\/$',
        })

        self.assertEqual(set(extract(self.urls1 + self.urls2,
                                     variableLengthFrags=True)), {
            r'^[a-z]{3,4}\.[a-z]{2,4}$',
            r'^[a-z]+\.com\/$',
            r'^[a-z]{3,4}[\.\/\:]{1,3}[a-z]+\.[a-z]{3,4}$',
            r'^http\:\/\/[a-z]+\.com\/$',
            r'^http\:\/\/www\.[a-z]+\.com\/$',
            r'^http\:\/\/www\.[a-z]+\.co\.uk\/$',
            r'^https?\:\/\/w{1,3}e?b?\.[a-z]+\.com$'
        })

    def test_agents(self):
        # This is really just a check that it doesn't bomb out, for the
        # moment. It's certainly not finding good REs at this point!
        examples_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                    'examples')
        agents_path = os.path.join(examples_dir, 'agents9.txt')
        with open(agents_path) as f:
            examples = f.read().splitlines()
        r = extract(examples, as_object=True)
        self.assertEqual(set(r.results.rex), {
            r'^Mozilla\/4\.0 \(compatible\; MSIE 8\.0\; Windows NT '
            r'6\.0\; Trident\/4\.0\;   Acoo Browser\; GTB5\; '
            r'Mozilla\/4\.0 \(compatible\; MSIE 6\.0\; Windows '
            r'NT 5\.1\;   SV1\) \; InfoPath\.1\; \.NET CLR '
            r'3\.5\.30729\; \.NET CLR 3\.0\.30618\)$',

            r'^Mozilla\/5\.0 \(compatible\; ABrowse 0\.4\; Syllable\)$',

            r'^Mozilla\/5\.0 \(compatible\; U\; ABrowse 0\.6\; '
            r'{1,2}Syllable\) AppleWebKit\/420\+ \(KHTML\, like Gecko\)$',

            r'^Mozilla\/5\.0 \(compatible\; MSIE 8\.0\; Windows '
            r'NT 6\.0\; Trident\/4\.0\; Acoo Browser 1\.98\.744\; '
            r'\.NET CLR {1,3}3\.5\.30729\)$',

            r'^Mozilla\/4\.0 \(compatible\; MSIE 7\.0\; Windows '
            r'NT 6\.0\; Acoo Browser\; SLCC1\;   \.NET CLR '
            r'2\.0\.50727\; Media Center PC 5\.0\; \.NET CLR '
            r'3\.0\.04506\)$',

            r'^Mozilla\/4\.0 \(compatible\; MSIE 8\.0\; Windows '
            r'NT 5\.1\; Trident\/4\.0\; SV1\; Acoo Browser\; '
            r'\.NET CLR 2\.0\.50727\; \.NET CLR 3\.0\.4506\.2152\; '
            r'\.NET CLR 3\.5\.30729\; Avant Browser\)$',

            r'^Mozilla\/4\.0 \(compatible\; MSIE 7\.0\; Windows '
            r'NT 6\.0\; Acoo Browser\; GTB5\; Mozilla\/4\.0 '
            r'\(compatible\; MSIE 6\.0\; Windows NT 5\.1\; '
            r'SV1\) \; Maxthon\; InfoPath\.1\; \.NET CLR 3\.5\.30729\; '
            r'\.NET CLR 3\.0\.30618\)$',
        })
        self.assertEqual(r.n_too_many_groups, 1)
        self.assertEqual(r.warnings[0], '1 string assigned to .{m,n} for '
                                        'needing "too many" groups.')

    def testmflag(self):
        patterns = ('a.1', 'b_2', 'c-3')
        expected = ['^c\-3$', r'^([a-z])([A-Z\.\_])(\d)$']
        r = extract(patterns, tag=True, extra_letters='._')
        self.assertEqual(r, expected)

    def testmflag2(self):
        patterns = ('a-1', 'c-3')
        expected = [r'^([a-z])\-(\d)$']
        r = extract(patterns, tag=True, extra_letters='._')
        self.assertEqual(r, expected)

    def testFindOuterCaptureGroups(self):
        r = re.compile('^(([A-Z]+) [1-5])([^\W_]+)$', re.U)
        m = re.match(r, 'THIS 3aéAB')
        self.assertEqual(is_outer_group(m, 1), True)
        self.assertEqual(is_outer_group(m, 2), False)
        self.assertEqual(is_outer_group(m, 3), True)

        f = group_map_function(m, 2)
        self.assertEqual(f(1), 1)
        self.assertEqual(f(2), 3)
        self.assertRaises(KeyError, f, 3)

        r = re.compile('^aaa([A-Z]+)bb$', re.U)
        m = re.match(r, 'aaaBBbb')
        f = group_map_function(m, 2)
        self.assertEqual(f(1), 1)
        self.assertRaises(KeyError, f, 2)

        regex = '@ (a(b)c) d (e(f(g)h(i(j)k)l)m) n (o) p (q(r)s) t'
        r = re.compile(regex, re.U)
        m = re.match(r, '@ abc d efghijklm n o p qrs t')
        f = group_map_function(m, 4)
        self.assertEqual(f(1), 1)
        self.assertEqual(f(2), 3)
        self.assertEqual(f(3), 8)
        self.assertEqual(f(4), 9)
        self.assertRaises(KeyError, f, 5)


    @unittest.skipIf(pandas is None, 'No pandas here')
    def testpdextract(self):
        df = pd.DataFrame({'a3': ["one", "two", pd.np.NaN],
                           'a45': ['three', 'four', 'five']})

        re3 = pdextract(df['a3'])
        re45 = pdextract(df['a45'])
        re345 = pdextract([df['a3'], df['a45']])

        self.assertEqual(re3, ['^[a-z]{3}$'])
        self.assertEqual(re45, ['^[a-z]{4,5}$'])
        self.assertEqual(re345, ['^[a-z]{3,5}$'])

    @unittest.skipIf(pandas is None, 'No pandas here')
    def testpdextract2(self):
        df = pd.DataFrame({'ab': ["one", True, pd.np.NaN]})
        self.assertRaisesRegex(ValueError, 'Non-null, non-string',
                               pdextract, df['ab'])


def print_ordered_dict(od):
    print()
    print('        expected = OrderedDict((')
    for k in od:
        print("            (u'%s',\n             rexpy.%s),"
              % (k, od[k]))
    print('        ))')


C = UNIC if UNICHRS else 'C'
def CtoUC(s):
    if UNICHRS:
        return s.replace('C', UNIC)
    else:
        return s



if sys.version_info[0] < 3:
    # Quieten down Python3's vexatious complaining
    TestExtraction.assertRaisesRegex = TestExtraction.assertRaisesRegexp

    # def testextractcli(self):
    #     examples_dir = os.path.join(os.path.abspath(__file__), 'examples')
    #     ids_path = os.path.join(examples_dir, 'ids.txt')
    #     params = get_params(ids_path)
    #     main(params)



if __name__ == '__main__':
    unittest.main()




# -*- coding: utf-8 -*-

"""
Test Suite
"""

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import datetime
import json
import math
import os
import time
import shutil
import subprocess
import sys
import tempfile
import unittest

from collections import OrderedDict
from distutils.spawn import find_executable

import pandas as pd
import numpy as np

from tdda.constraints.base import (
    MinConstraint,
    MaxConstraint,
    SignConstraint,
    TypeConstraint,
    MaxNullsConstraint,
    NoDuplicatesConstraint,
    AllowedValuesConstraint,
    MinLengthConstraint,
    MaxLengthConstraint,
    DatasetConstraints,
    Fields,
    FieldConstraints,
    verify,
    native_definite,
    UTF8DefiniteObject,
    fuzzy_less_than,
    fuzzy_greater_than,
)
from tdda.constraints.console import main_with_argv

from tdda.constraints.pd import constraints as pdc
from tdda.constraints.pd.constraints import (load_df, verify_df,
                                             discover_df, detect_df)
from tdda.constraints.pd.discover import discover_df_from_file
from tdda.constraints.pd.verify import verify_df_from_file

from tdda.referencetest import ReferenceTestCase

isPython2 = sys.version_info[0] < 3

THIS_DIR = os.path.dirname(os.path.abspath(__file__))
TESTDATA_DIR = os.path.join(os.path.dirname(THIS_DIR), 'testdata')


SMALL = 2.48e-324
MILLION = 1000 * 1000
REAL_MILLION = 1000 * 1000.0
BOOLS = (True, False)
SMI = 9223372036854775807
POS_INTS = (1, SMI, SMI + 1)
NEG_INTS = (-1, -SMI - 1, -SMI - 2)
INTS = POS_INTS + (0,) + NEG_INTS
POS_REALS = (1.0, SMALL, SMI * math.pi, float('inf'), time.time())
NEG_REALS = (-1.0, -SMALL, -SMI * math.pi, -float('inf'), -time.time())
REALS = POS_REALS + (0.0,) + NEG_REALS
NUMBERS = BOOLS + INTS + REALS
STRINGS = ('', 'a', 'αβγδε')
NULLS = (None,)
DATES = (datetime.datetime(1970, 1, 1),
         datetime.datetime(1, 1, 1),
         datetime.datetime(9999, 12, 31, 23, 59, 59),
         datetime.datetime.now(),
         datetime.datetime.utcnow())
OTHERS = (3 + 4j, lambda x: 1, [], (), {}, Exception) + ((u'u',) if isPython2
                                                              else (b'u',))


THIS_DIR = os.path.dirname(os.path.abspath(__file__))
TESTDATA_DIR = os.path.join(os.path.dirname(THIS_DIR), 'testdata')


class ConstraintVerificationTester:
    """
    Delegate class for checking constraint verifications.

    This exists so that when there are test failures,
    useful information about failing inputs and outputs can be shown.
    """
    outstandingAssertions = 0
    def __init__(self, tester, *args, **kwargs):
        self.tester = tester
        self.verifier = pdc.PandasConstraintVerifier(*args, **kwargs)

    def __getattr__(self, k):
        if k == 'decOutstanding':
            return self.decOutstanding
        return lambda *args, **kwargs: self.invoke(k, *args, **kwargs)

    def invoke(self, method, *args, **kwargs):
        satisfied = getattr(self.verifier, method)(*args, **kwargs)
        ConstraintVerificationTester.outstandingAssertions += 1
        return Asserter(self, satisfied, method, args, kwargs)

    def decOutstanding(self):
        ConstraintVerificationTester.outstandingAssertions -= 1


class Asserter:
    """
    """
    def __init__(self, cvt, satisfied, method, args, kwargs):
        self.cvt = cvt
        self.satisfied = satisfied
        self.method = method
        self.args = args
        self.kwargs = kwargs

    def isTrue(self):
        self.cvt.decOutstanding()
        self.cvt.tester.assertTrue(self.satisfied, self.diagnostic(True))

    def isFalse(self):
        self.cvt.decOutstanding()
        self.cvt.tester.assertFalse(self.satisfied, self.diagnostic(False))

    def diagnostic(self, expected):
        return ('Verifier: %s Inputs: %s %s: Assertion: %s'
                % (self.method, str(self.args),
                   str(self.kwargs) if self.kwargs else '',
                   'satisified' if expected is True
                                else 'not satisfied' if expected is False
                                else expected))



class TestPandasConstraintVerifiers(ReferenceTestCase):

    @classmethod
    def tearDownClass(cls):
        assert ConstraintVerificationTester.outstandingAssertions == 0

    def test_tdda_types_of_base_types(self):
        self.assertEqual(pdc.pandas_tdda_type(None), 'null')
        for v in BOOLS:
            self.assertEqual(pdc.pandas_tdda_type(v), 'bool')
        for v in INTS:
            self.assertEqual(pdc.pandas_tdda_type(v), 'int')
        for v in REALS:
            self.assertEqual(pdc.pandas_tdda_type(v), 'real')
        for v in STRINGS:
            self.assertEqual(pdc.pandas_tdda_type(v), 'string')
        for v in DATES:
            self.assertEqual(pdc.pandas_tdda_type(v), 'date')
        for v in OTHERS:
            self.assertEqual(pdc.pandas_tdda_type(v), 'other')

    def test_coarse_types_of_base_types(self):
        self.assertEqual(pdc.pandas_tdda_type(None), 'null')
        for v in BOOLS:
            self.assertEqual(pdc.pandas_coarse_type(v), 'number')
        for v in INTS:
            self.assertEqual(pdc.pandas_coarse_type(v), 'number')
        for v in REALS:
            self.assertEqual(pdc.pandas_coarse_type(v), 'number')
        for v in STRINGS:
            self.assertEqual(pdc.pandas_coarse_type(v), 'string')
        for v in DATES:
            self.assertEqual(pdc.pandas_coarse_type(v), 'date')
        for v in OTHERS:
            self.assertEqual(pdc.pandas_coarse_type(v), 'other')

    def test_compatibility(self):
        for kind in (NUMBERS, STRINGS, DATES):
            x = kind[0]
            y = kind[-1]
            self.assertTrue(pdc.pandas_types_compatible(x, y))
            self.assertTrue(pdc.pandas_types_compatible(x, x))

    def test_incompatibility(self):
        for X in (NUMBERS, STRINGS, DATES, OTHERS, NULLS):
            for Y in (NUMBERS, STRINGS, DATES, OTHERS, NULLS):
                if X is not Y:
                    self.assertFalse(pdc.pandas_types_compatible(X[0], Y[0]))
                    self.assertFalse(pdc.pandas_types_compatible(Y[0], X[0]))

    def test_fuzzy_less_than_zero(self):
        verifier = pdc.PandasConstraintVerifier(df=None)
        epsilon = 0.01
        for x in NEG_REALS:
            self.assertTrue(fuzzy_less_than(x, 0.0, epsilon))
            self.assertFalse(fuzzy_less_than(0.0, x, epsilon))
        for x in POS_REALS:
            self.assertFalse(fuzzy_less_than(x, 0.0, epsilon))
            self.assertTrue(fuzzy_less_than(0.0, x, epsilon))

        self.assertTrue(fuzzy_less_than(0.0, 0.0, epsilon))
        self.assertTrue(fuzzy_less_than(0.0, SMALL / 2, epsilon))  # == 0.0
        self.assertEqual(SMALL / 2, 0.0)

    def test_fuzzy_less_than(self):
        goods = (
            (1.0, 1.0),
            (1.00999, 1.0),

            (MILLION, MILLION),
            (MILLION + 9999, MILLION),
            (MILLION + 10000, MILLION),
            (REAL_MILLION, REAL_MILLION),
            (MILLION + 9999.0, REAL_MILLION),
            (MILLION + 10000.0, MILLION),

            (1e-100, 1e-100),
            (1.00999e-100, 1e-100),

            (-1.0, -1.0),
            (-.990001, -1.0),

            (-MILLION, -MILLION),
            (-MILLION + 9999, -MILLION),
            (-MILLION + 10000, -MILLION),
            (-REAL_MILLION, -REAL_MILLION),
            (-REAL_MILLION + 9999.0, -REAL_MILLION),
            (-REAL_MILLION + 10000.0, -MILLION),

            (-1e-100, -1e-100),
            (-0.999999e-100, -1e-100),
        )
        bad_goods = (
            (MILLION + 10000 + SMALL, MILLION),  # Should fail mathematically.
                                                 # But rounding
        )
        bads = (
            (1.01000001, 1.0),
            (MILLION + 10001, MILLION),
            (MILLION + 10000.0000000001, MILLION),
            (1.010001e-100, 1.0e-100),

            (-0.989999, -1.0),
            (-MILLION + 100001, -MILLION),
            (-MILLION + 10000.0000000001, -MILLION),
            (-0.98999e-100, -1.0e-100),
        )

        cvt = ConstraintVerificationTester(self, df=None)
        self.assertEqual(MILLION + 10000 + SMALL, MILLION + 10000)
        epsilon = 0.01
        for (x, y) in goods + bad_goods:
            self.assertTrue(fuzzy_less_than(x, y, epsilon))
        for (x, y) in bads:
            self.assertFalse(fuzzy_less_than(x, y, epsilon))

    def test_fuzzy_greater_than_zero(self):
        cvt = ConstraintVerificationTester(self, df=None)
        epsilon = 0.01
        for x in POS_REALS:
            self.assertTrue(fuzzy_greater_than(x, 0.0, epsilon))
            self.assertFalse(fuzzy_greater_than(0.0, x, epsilon))
        for x in NEG_REALS:
            self.assertFalse(fuzzy_greater_than(x, 0.0, epsilon))
            self.assertTrue(fuzzy_greater_than(0.0, x, epsilon))

        self.assertTrue(fuzzy_greater_than(0.0, 0.0, epsilon))
        self.assertTrue(fuzzy_greater_than(0.0, SMALL / 2, epsilon))  # == 0.0
        self.assertEqual(SMALL / 2, 0.0)

    def test_fuzzy_greater_than(self):
        goods = (
            (1.0, 1.0),
            (.9900001, 1.0),

            (MILLION, MILLION),
            (999900, MILLION),
            (999899, MILLION),
            (999899.999, REAL_MILLION),

            (1e-100, 1e-100),
            (0.99000001e-100, 1e-100),

            (-1.0, -1.0),
            (-1.009999, -1.0),

            (-MILLION, -MILLION),
            (-MILLION + 10000, -MILLION),
            (-REAL_MILLION, -REAL_MILLION),
            (-REAL_MILLION + 10000.0001, -REAL_MILLION),
            (-REAL_MILLION + 10000.0, -MILLION),

            (-1e-100, -1e-100),
            (-0.99001e-100, -1e-100),
        )
        bad_goods = (
            (999900 - SMALL, MILLION),  # Should fail mathematically.
                                        # But MILLION + SMALL == MILLION
        )
        bads = (
            (0.9899999, 1.0),
            (MILLION - 10001, MILLION),
            (MILLION - 10000.0000000001, MILLION),
            (0.989999-100, 1.0e-100),

            (-1.01001, -1.0),
            (-MILLION - 10001, -MILLION),
            (-MILLION - 10000.0000000001, -MILLION),
            (-1.0100001e-100, -1.0e-100),

        )
        cvt = ConstraintVerificationTester(self, df=None)
        self.assertEqual(999900 - SMALL, 999900)
        epsilon = 0.01
        for (x, y) in goods + bad_goods:
            self.assertTrue(fuzzy_greater_than(x, y, epsilon))
        for (x, y) in bads:
            self.assertFalse(fuzzy_greater_than(x, y, epsilon))

    def test_caching(self):
        df = pd.DataFrame({
            'a': range(3),
        })
        v = pdc.PandasConstraintVerifier(df)

        # First check the max gets computed and cached correctly
        self.assertEqual(v.get_max('a'), 2)
        self.assertEqual(v.cache['a']['max'], 2)

        # write new results (which are obviously wrong, but no matter)
        v.cache['a']['max'] = -3

        # Now check the new (wrong) values are returned, and remain cached
        self.assertEqual(v.get_max('a'), -3)
        self.assertEqual(v.cache['a']['max'], -3)

    def test_verify_min_constraint(self):
        df = pd.DataFrame({
            'intzero': range(3),
            'intzeron': [0, None, 1],
            'intzeronn': [0, np.nan, 1],

            'realzero': [float(i) for i in range(3)],
            'realzeron': [0.0, None, 1.0],
            'realzeronn': [0.0, np.nan, 1.0],

            'int1': range(1, 4),
            'real1': [float(i) for i in range(1, 4)],

            'sempty': [''] * 3,
            'semptyn': ['', '', None],

            'sabc': ['a', 'b', 'c'],
            'sacn': ['a', 'c', None],

        })
        goods0 = [
            (col, v, p)
            for p in (None, 'closed', 'fuzzy')
            for v in (0, 0.0)
            for col in ('intzero', 'intzeron', 'intzeronn',
                        'realzero', 'realzeron', 'realzeronn')
        ]
        bads0 = [
            (col, v, 'open')
            for v in (0, 0.0)
            for col in ('intzero', 'intzeron', 'intzeronn',
                        'realzero', 'realzeron', 'realzeronn')
        ]

        goods1 = [
            (col, v, p)
            for p in (None, 'closed', 'fuzzy')
            for v in (1, 1.0)
            for col in ('int1', 'real1')
        ]
        bads1 = [
            (col, v, 'open')
            for v in (1, 1.0)
            for col in ('int1', 'real1')
        ]

        good_strings = [
            ('sempty', '', 'closed'),
            ('semptyn', '', 'closed'),
            ('sabc', '', 'closed'),
            ('sacn', '', 'closed'),
            ('sabc', 'a', 'closed'),
            ('sacn', 'a', 'closed'),
            ('sabc', 'A', 'closed'),
            ('sacn', 'A', 'closed'),
        ]

        bad_strings = [
            ('sempty', 'a', 'closed'),
            ('semptyn', 'A', 'closed'),
            ('sabc', 'b', 'closed'),
            ('sacn', 'b', 'closed'),
            ('sabc', 'foo', 'closed'),
            ('sacn', 'bar', 'closed'),
        ]

        goods = goods0 + goods1 + good_strings
        bads = bads0 + bads1 + bad_strings
        cvt = ConstraintVerificationTester(self, df)
        for (col, value, precision) in goods:
            c = MinConstraint(value, precision=precision)
            cvt.verify_min_constraint(col, c).isTrue()
        for (col, value, precision) in bads:
            c = MinConstraint(value, precision=precision)
            cvt.verify_min_constraint(col, c).isFalse()

    def test_verify_max_constraint(self):
        df = pd.DataFrame({
            'intzero': range(-2, 1),
            'intzeron': [0, None, -1],
            'intzeronn': [0, np.nan, -1],

            'realzero': [-float(i) for i in range(3)],
            'realzeron': [0.0, None, -1.0],
            'realzeronn': [0.0, np.nan, -1.0],

            'int1': range(-1, 2),
            'real1': [float(i) for i in range(-1, 2)]

        })
        goods0 = [
            (col, v, p)
            for p in (None, 'closed', 'fuzzy')
            for v in (0, 0.0)
            for col in ('intzero', 'intzeron', 'intzeronn',
                        'realzero', 'realzeron', 'realzeronn')
        ]
        bads0 = [
            (col, v, 'open')
            for v in (0, 0.0)
            for col in ('intzero', 'intzeron', 'intzeronn',
                        'realzero', 'realzeron', 'realzeronn')
        ]
        goods1 = [
            (col, v, p)
            for p in (None, 'closed', 'fuzzy')
            for v in (1, 1.0)
            for col in ('int1', 'real1')
        ]
        bads1 = [
            (col, v, 'open')
            for v in (1, 1.0)
            for col in ('int1', 'real1')
        ]

        goods = goods0 + goods1
        bads = bads0 + bads1

        cvt = ConstraintVerificationTester(self, df)
        for (col, value, precision) in goods:
            c = MaxConstraint(value, precision=precision)
            cvt.verify_max_constraint(col, c).isTrue()
        for (col, value, precision) in bads:
            c = MaxConstraint(value, precision=precision)
            cvt.verify_max_constraint(col, c).isFalse()

    def test_verify_min_max_length_constraints(self):
        df = pd.DataFrame({
            'zero': [''] * 4,
            'zeroOne': ['', 'a', '1', None],
            'one': ['α', 'b', 'c', None],       # Note unicode; min max len 1
            'oneTwo': ['a', 'aa', 'bb', None],
            'two': ['αα', 'αα', 'ββ', 'ββ'],    # Note unicode; min max len 2
        })
        goods = [
            ('zero', 0, 0),
            ('zero', 0, 10),

            ('zeroOne', 0, 1),
            ('zeroOne', 0, 5),

            ('one', 1, 1),
            ('one', 0, 1),
            ('one', 1, 4),
            ('one', 0, 10),

            ('oneTwo', 1, 2),
            ('oneTwo', 0, 2),
            ('oneTwo', 1, 4),
            ('oneTwo', 0, 10),

            ('two', 2, 2),
            ('two', 0, 2),
            ('two', 2, 8),
            ('two', 0, 7),
        ]
        bads = [
            ('zero', 1, None),
            ('zeroOne', 2, 0),
            ('one', 2, 0),
            ('oneTwo', 3, 0),
            ('two', 3, 1),
        ]
        cvt = ConstraintVerificationTester(self, df)
        for (col, m, M) in goods:
            c = MinLengthConstraint(m)
            cvt.verify_min_length_constraint(col, c).isTrue()
            c = MaxLengthConstraint(M)
            cvt.verify_max_length_constraint(col, c).isTrue()
        for (col, m, M) in bads:
            c = MinLengthConstraint(m)
            cvt.verify_min_length_constraint(col, c).isFalse()
            if M is not None:
                c = MaxLengthConstraint(M)
                cvt.verify_max_length_constraint(col, c).isFalse()

    def test_verify_tdda_type_constraint(self):
        df = pd.DataFrame({
            'b': [True, False],
            'i': [1, 0],
            'r': [1.0, 1.1],
            's': ['1', 'a'],
            'd': [datetime.datetime(2000,1,1), datetime.datetime(2000,1,2)],

            'bn': [True, None],
            'in': [1, None],
            'rn': [1.1, None],
            'sn': [None, 'a'],
            'dn': [datetime.datetime(2000,1,1), None],
        })
        goods = [
            ('b', 'bool', 'strict'),
            ('i', 'int', 'strict'),
            ('r', 'real', 'strict'),
            ('s', 'string', 'strict'),
            ('d', 'date', 'strict'),

            ('bn', 'bool', 'sloppy'), # Fails strict because promoted to object
            ('in', 'int', 'sloppy'),  # Fails strict because promoted to real
            ('rn', 'real', 'strict'),
            ('sn', 'string', 'strict'),
            ('dn', 'date', 'strict'),

            ('bn', ['bool', 'string'], 'strict'),   # promotion to object
            ('in', ['int', 'real'], 'strict'),      # promotion to real
            ('rn', ['int', 'real'], 'strict'),      # just a looser constraint
        ]
        bads = [
            ('bn', 'bool', 'strict'),
            ('in', 'int', 'strict'),
            ('b', 'date', 'sloppy'),
            ('i', 'bool', 'sloppy'),
            ('r', 'int', 'sloppy'),
            ('s', 'real', 'sloppy'),
            ('d', 'string', 'sloppy'),

            ('s', ['real', 'int'], 'sloppy'),  # not sloppy enough for this
        ]
        strict_cvt = ConstraintVerificationTester(self, df,
                                                  type_checking='strict')
        sloppy_cvt = ConstraintVerificationTester(self, df,
                                                  type_checking='sloppy')
        for (col, value, strictness) in goods:
            cvt = strict_cvt if strictness == 'strict' else sloppy_cvt
            c = TypeConstraint(value)
            if strictness == 'strict':
                strict_cvt.verify_tdda_type_constraint(col, c).isTrue()
            # Any field that satisfies the strict checker should
            # also satisfy the sloppy checker
            sloppy_cvt.verify_tdda_type_constraint(col, c).isTrue()

        for (col, value, strictness) in bads:
            c = TypeConstraint(value)
            if strictness == 'sloppy':
                sloppy_cvt.verify_tdda_type_constraint(col, c).isFalse()
            # Any field that fails to satisfy the sloppy checker should
            # also fail to satisfy the strict checker
            strict_cvt.verify_tdda_type_constraint(col, c).isFalse()

    def test_verify_tdda_sign_constraint(self):
        df = pd.DataFrame({
            'ipos': [1, 2, 3],
            'inonneg': [0, 1, 2],
            'izero': [0, 0, 0],
            'inonpos': [0, -1, -2],
            'ineg': [-1, -2, -3],
            'imixed': [1,0,-1],

            'rpos': [1.0, 2.0, 3.0],
            'rnonneg': [0.0, 1.0, 1.0],
            'rzero': [0.0, 0.0, 0.0],
            'rnonpos': [0.0, -1.0, -1.0],
            'rneg': [-1.0, -2.0, -3.0],
            'rmixed': [-1.0, 0.0, 1.0],

            'iposn': [1, 2, np.nan],
            'inonnegn': [0, 1, np.nan],
            'izeron': [0, 0, np.nan],
            'inonposn': [0, -1, np.nan],
            'inegn': [-1, -2, np.nan],
            'inmixedn': [-1, 1, np.nan],

            'rposn': [1.0, 2.0, np.nan],
            'rnonnegn': [0.0, 1.0, np.nan],
            'rzeron': [0.0, 0.0, np.nan],
            'rnonposn': [0.0, -1.0, np.nan],
            'rnegn': [-1.0, -2.0, np.nan],
            'rnmixedn': [-1.0, 1.0, np.nan],

            'null': [np.nan, np.nan, np.nan],
        })
        cvt = ConstraintVerificationTester(self, df)
        for col in df:
            pos = SignConstraint('positive')
            nonneg = SignConstraint('non-negative')
            zero = SignConstraint('zero')
            nonpos = SignConstraint('non-positive')
            neg = SignConstraint('negative')
            null = SignConstraint('null')
            if 'nonpos' in col:
                cvt.verify_sign_constraint(col, pos).isFalse()
                cvt.verify_sign_constraint(col, nonneg).isFalse()
                cvt.verify_sign_constraint(col, zero).isFalse()
                cvt.verify_sign_constraint(col, nonpos).isTrue()
                cvt.verify_sign_constraint(col, neg).isFalse()
                cvt.verify_sign_constraint(col, null).isFalse()
            elif 'pos' in col:
                cvt.verify_sign_constraint(col, pos).isTrue()
                cvt.verify_sign_constraint(col, nonneg).isTrue()
                cvt.verify_sign_constraint(col, zero).isFalse()
                cvt.verify_sign_constraint(col, nonpos).isFalse()
                cvt.verify_sign_constraint(col, neg).isFalse()
                cvt.verify_sign_constraint(col, null).isFalse()
            elif 'nonneg' in col:
                cvt.verify_sign_constraint(col, pos).isFalse()
                cvt.verify_sign_constraint(col, nonneg).isTrue()
                cvt.verify_sign_constraint(col, zero).isFalse()
                cvt.verify_sign_constraint(col, nonpos).isFalse()
                cvt.verify_sign_constraint(col, neg).isFalse()
                cvt.verify_sign_constraint(col, null).isFalse()
            elif 'neg' in col:
                cvt.verify_sign_constraint(col, pos).isFalse()
                cvt.verify_sign_constraint(col, nonneg).isFalse()
                cvt.verify_sign_constraint(col, zero).isFalse()
                cvt.verify_sign_constraint(col, nonpos).isTrue()
                cvt.verify_sign_constraint(col, neg).isTrue()
                cvt.verify_sign_constraint(col, null).isFalse()
            elif 'zero' in col:
                cvt.verify_sign_constraint(col, pos).isFalse()
                cvt.verify_sign_constraint(col, nonneg).isTrue()
                cvt.verify_sign_constraint(col, zero).isTrue()
                cvt.verify_sign_constraint(col, nonpos).isTrue()
                cvt.verify_sign_constraint(col, neg).isFalse()
                cvt.verify_sign_constraint(col, null).isFalse()
            elif col == 'null':
                pass
                cvt.verify_sign_constraint(col, pos).isTrue()
                cvt.verify_sign_constraint(col, nonneg).isTrue()
                cvt.verify_sign_constraint(col, zero).isTrue()
                cvt.verify_sign_constraint(col, nonpos).isTrue()
                cvt.verify_sign_constraint(col, neg).isTrue()
                cvt.verify_sign_constraint(col, null).isTrue()
            elif 'mixed' in col:
                cvt.verify_sign_constraint(col, pos).isFalse()
                cvt.verify_sign_constraint(col, nonneg).isFalse()
                cvt.verify_sign_constraint(col, zero).isFalse()
                cvt.verify_sign_constraint(col, nonpos).isFalse()
                cvt.verify_sign_constraint(col, neg).isFalse()
                cvt.verify_sign_constraint(col, null).isFalse()
            else:
                raise Exception('Cannot get here: %s' % col)

    def test_verify_tdda_max_nulls_constraint(self):
        df = pd.DataFrame({
            'b': [True, False],
            'i': [1, -1],
            'r': [1.0, -1.0],
            's': ['a', 'a'],
            'd': [datetime.datetime.now()] * 2,

            'bn': [True, None],
            'in': [1, None],
            'rn': [None, 1.0],
            'sn': [None, 'a'],
            'dn': [datetime.datetime.now(), None],

            'n': [None, None],
        })
        cvt = ConstraintVerificationTester(self, df)
        c = MaxNullsConstraint(0)
        for col in df:
            if col.endswith('n'):
                 cvt.verify_max_nulls_constraint(col, c).isFalse()
            else:
                 cvt.verify_max_nulls_constraint(col, c).isTrue()

    def test_verify_no_duplicates_constraint(self):
        df = pd.DataFrame({
            'bu': [True, False, None, None],  # Note two nulls
            'iu': [1, -1, 0, 2],
            'ru': [1.0, -1.0, 0.0, 3.0],
            'su': ['a', 'b', 'c', ''],
            'du': [datetime.datetime(2000,1,1), datetime.datetime(2000,1,2),
                   datetime.datetime(2000,1,3), datetime.datetime(2000,1,4)],

            'bd': [True, True, False, None],
            'id': [1, 2, 2, 3],
            'rd': [1.0, 2.0, 2.0, 3.0],
            'sd': ['a', 'a', 'a', 'a'],
            'dd': [datetime.datetime(2000,1,1)] * 4,

            'IU': [1, -1, None, None],
            'RU': [1.0, -1.0, None, None],
            'SU': [None, None, 'c', ''],
            'DU': [datetime.datetime(2000,1,1), None,
                   None, datetime.datetime(2000,1,4)],

            'BD': [True, True, False, None],
            'ID': [1, 2, 2, None],
            'RD': [None, 2.0, 2.0, None],
            'SD': ['a', 'a', None, 'b'],
            'DD': [datetime.datetime(2000,1,1)] * 2 + [None] * 2,

            'nu': [None, None, None, None],
        })
        cvt = ConstraintVerificationTester(self, df)
        c = NoDuplicatesConstraint(True)
        for col in df:
            if col.lower().endswith('u'):
                 cvt.verify_no_duplicates_constraint(col, c).isTrue()
            else:
                 cvt.verify_no_duplicates_constraint(col, c).isFalse()

    def test_verify_allowed_values_constraint(self):
        DIGITS = list('1234567890')
        PRIMES = list('2357')
        EMPTIES = ['', ' ', '  ', '   ']
        ANDOR = ['and', 'or']
        df = pd.DataFrame({
            'digits': list('8275'),
            'primes1': PRIMES,
            'primes2': list('3355'),
            'empties': EMPTIES,
            'eitheror': ['and', 'not', 'either', 'or'],

            'digitsn': list('827') + [None],
            'primes1n': PRIMES[:-1] + [np.nan],
            'primes2n': [None, np.nan, '3', '5'],
            'emptiesn': [' ', None, '  ', '   '],
            'eitherorn': ['and', None, 'either', 'or'],

            'null': [None] * 4,
        })
        cvt = ConstraintVerificationTester(self, df)
        c_digits = AllowedValuesConstraint(DIGITS)
        c_primes = AllowedValuesConstraint(PRIMES)
        c_empties = AllowedValuesConstraint(EMPTIES)
        c_andor = AllowedValuesConstraint(ANDOR)
        c_nothing = AllowedValuesConstraint([])

        cvt.verify_allowed_values_constraint('digits', c_digits).isTrue()
        cvt.verify_allowed_values_constraint('digitsn', c_digits).isTrue()
        cvt.verify_allowed_values_constraint('primes1', c_primes).isTrue()
        cvt.verify_allowed_values_constraint('primes1n', c_primes).isTrue()
        cvt.verify_allowed_values_constraint('primes2', c_primes).isTrue()
        cvt.verify_allowed_values_constraint('primes2n', c_primes).isTrue()
        cvt.verify_allowed_values_constraint('primes1', c_digits).isTrue()
        cvt.verify_allowed_values_constraint('primes1n', c_digits).isTrue()
        cvt.verify_allowed_values_constraint('primes2', c_digits).isTrue()
        cvt.verify_allowed_values_constraint('primes2n', c_digits).isTrue()
        cvt.verify_allowed_values_constraint('empties', c_empties).isTrue()
        cvt.verify_allowed_values_constraint('emptiesn', c_empties).isTrue()

        cvt.verify_allowed_values_constraint('eitheror', c_empties).isFalse()
        cvt.verify_allowed_values_constraint('eitherorn', c_empties).isFalse()

        cvt.verify_allowed_values_constraint('digits', c_primes).isFalse()
        cvt.verify_allowed_values_constraint('digitsn', c_primes).isFalse()
        cvt.verify_allowed_values_constraint('primes1', c_empties).isFalse()
        cvt.verify_allowed_values_constraint('primes1n', c_andor).isFalse()
        cvt.verify_allowed_values_constraint('primes2', c_andor).isFalse()
        cvt.verify_allowed_values_constraint('primes2n', c_empties).isFalse()
        cvt.verify_allowed_values_constraint('empties', c_primes).isFalse()
        cvt.verify_allowed_values_constraint('emptiesn', c_digits).isFalse()

        for col in df:
            if col == 'null':
                cvt.verify_allowed_values_constraint(col, c_nothing).isTrue()
            else:
                cvt.verify_allowed_values_constraint(col, c_nothing).isFalse()

    def testFieldVerification(self):
        df1 = pd.DataFrame({
                'b': [True, False] * 2,
                'i': range(1, 5),
                'r': [float(x) for x in range(1, 5)],
                's': ['S%s' % x for x in range(1, 5)],
                'd': [datetime.datetime(2016,1,x) for x in range(1, 5)]
             })
        ic1 = FieldConstraints('i',
                               [TypeConstraint('int'),
                                MinConstraint(0),
                                MaxConstraint(10),
                                SignConstraint('positive'),
                                MaxNullsConstraint(0),
                                NoDuplicatesConstraint()])

        ic2 = FieldConstraints('i',
                               [TypeConstraint('bool'),
                                MinConstraint(2),
                                MaxConstraint(3),
                                SignConstraint('negative'),
                                MaxNullsConstraint(0),
                                NoDuplicatesConstraint()])

        dfc1 = [ic1]
        dsc1 = DatasetConstraints(dfc1)
        pdcv1 = pdc.PandasConstraintVerifier(df1)
        results1 = verify(dsc1, list(df1), pdcv1.verifiers())
        expected = ('FIELDS:\n\n'
                    'i: 0 failures  6 passes  '
                    'type ✓  min ✓  max ✓  sign ✓  '
                    'max_nulls ✓  no_duplicates ✓\n\n'
                    'SUMMARY:\n\nConstraints passing: 6\nConstraints failing: 0')
        self.assertEqual(str(results1), expected)
        expected = pd.DataFrame(OrderedDict((
                        ('field', ['i']),
                        ('failures', [0]),
                        ('passes', [6]),
                        ('type', [True]),
                        ('min', [True]),
                        ('max', [True]),
                        ('sign', [True]),
                        ('max_nulls', [True]),
                        ('no_duplicates', [True]),
                        )))
        vdf = pdc.PandasVerification.verification_to_dataframe(results1)
        self.assertTrue(vdf.equals(expected))

        df2 = pd.DataFrame({'i': [1, 2, 2, 6, np.nan]})
        dfc2 = [ic2]
        dsc2 = DatasetConstraints(dfc2)
        pdcv2 = pdc.PandasConstraintVerifier(df2)
        results2 = verify(dsc2, list(df2), pdcv2.verifiers())
        # expect the boolean->real type constraint to pass with sloppy types
        expected = ('FIELDS:\n\n'
                    'i: 5 failures  1 pass  '
                    'type ✓  min ✗  max ✗  sign ✗  '
                    'max_nulls ✗  no_duplicates ✗\n\n'
                    'SUMMARY:\n\nConstraints passing: 1\nConstraints failing: 5')
        self.assertEqual(str(results2), expected)
        expected = pd.DataFrame(OrderedDict((
                        ('field', ['i']),
                        ('failures', [5]),
                        ('passes', [1]),
                        ('type', [True]),
                        ('min', [False]),
                        ('max', [False]),
                        ('sign', [False]),
                        ('max_nulls', [False]),
                        ('no_duplicates', [False]),
                        )))
        vdf = pdc.PandasVerification.verification_to_dataframe(results2)
        self.assertTrue(vdf.equals(expected))

        pdcv2strict = pdc.PandasConstraintVerifier(df2, type_checking='strict')
        results2strict = verify(dsc2, list(df2), pdcv2strict.verifiers())
        # expect the boolean->real type constraint to fail with strict types
        expected = ('FIELDS:\n\n'
                    'i: 6 failures  0 passes  '
                    'type ✗  min ✗  max ✗  sign ✗  '
                    'max_nulls ✗  no_duplicates ✗\n\n'
                    'SUMMARY:\n\nConstraints passing: 0\nConstraints failing: 6')
        self.assertEqual(str(results2strict), expected)
        expected = pd.DataFrame(OrderedDict((
                        ('field', ['i']),
                        ('failures', [6]),
                        ('passes', [0]),
                        ('type', [False]),
                        ('min', [False]),
                        ('max', [False]),
                        ('sign', [False]),
                        ('max_nulls', [False]),
                        ('no_duplicates', [False]),
                        )))
        vdf = pdc.PandasVerification.verification_to_dataframe(results2strict)
        self.assertTrue(vdf.equals(expected))

        ic3 = FieldConstraints('i', [TypeConstraint('int')])
        df3 = df1
        dfc3 = [ic3]
        dsc3 = DatasetConstraints(dfc3)
        pdcv3 = pdc.PandasConstraintVerifier(df3)
        results3 = verify(dsc3, list(df3), pdcv3.verifiers())
        expected = ('FIELDS:\n\n'
                    'i: 0 failures  1 pass  type ✓\n\n'
                    'SUMMARY:\n\nConstraints passing: 1\nConstraints failing: 0')
        self.assertEqual(str(results3), expected)
        expected = pd.DataFrame(OrderedDict((
                        ('field', ['i']),
                        ('failures', [0]),
                        ('passes', [1]),
                        ('type', [True]),
                   )))
        vdf = pdc.PandasVerification.verification_to_dataframe(results3)
        self.assertTrue(vdf.equals(expected))

        pdcv3 = pdc.PandasConstraintVerifier(df3)
        results3 = verify(dsc3, list(df3), pdcv3.verifiers(), ascii=True)
        expected = ('FIELDS:\n\n'
                    'i: 0 failures  1 pass  type OK\n\n'
                    'SUMMARY:\n\nConstraints passing: 1\nConstraints failing: 0')
        self.assertEqual(str(results3), expected)

    def testElements92(self):
        csv_path = os.path.join(TESTDATA_DIR, 'elements92.csv')
        df = pd.read_csv(csv_path)
        constraints_path = os.path.join(TESTDATA_DIR, 'elements92.tdda')
        v = verify_df(df, constraints_path)
        self.assertEqual(v.passes, 72)
        self.assertEqual(v.failures, 0)

    def testElements92rex(self):
        csv_path = os.path.join(TESTDATA_DIR, 'elements92.csv')
        df = pd.read_csv(csv_path)
        constraints_path = os.path.join(TESTDATA_DIR, 'elements92rex.tdda')
        v = verify_df(df, constraints_path)
        self.assertEqual(v.passes, 78)
        self.assertEqual(v.failures, 0)

    def testElements118(self):
        csv_path = os.path.join(TESTDATA_DIR, 'elements118.csv')
        df = pd.read_csv(csv_path)
        constraints_path = os.path.join(TESTDATA_DIR, 'elements92.tdda')
        v = verify_df(df, constraints_path, report='fields')
        self.assertEqual(v.passes, 57)
        self.assertEqual(v.failures, 15)
        vdf = v.to_dataframe()
        vdf.sort_values('field', inplace=True)
        self.assertStringCorrect(vdf.to_string(), 'elements118.df')

    def testElements118rex(self):
        csv_path = os.path.join(TESTDATA_DIR, 'elements118.csv')
        df = pd.read_csv(csv_path)
        constraints_path = os.path.join(TESTDATA_DIR, 'elements92rex.tdda')
        v = verify_df(df, constraints_path, report='fields')
        self.assertEqual(v.passes, 61)
        self.assertEqual(v.failures, 17)
        vdf = v.to_dataframe()
        vdf.sort_values('field', inplace=True)
        self.assertStringCorrect(vdf.to_string(), 'elements118rex.df')

    def testDetectElements118rexToFile(self):
        csv_path = os.path.join(TESTDATA_DIR, 'elements118.csv')
        df = pd.read_csv(csv_path)
        constraints_path = os.path.join(TESTDATA_DIR, 'elements92rex.tdda')
        detectfile = os.path.join(self.tmp_dir, 'elements118rex_detect.csv')
        v = detect_df(df, constraints_path, report='fields',
                      outpath=detectfile, output_fields=['Z'],
                      rowindex_is_index=False)
        self.assertEqual(v.passes, 61)
        self.assertEqual(v.failures, 17)
        self.assertFileCorrect(detectfile, 'elements118rex_detect.csv')

    def testDetectElements118rexToFilePerConstraint(self):
        csv_path = os.path.join(TESTDATA_DIR, 'elements118.csv')
        df = pd.read_csv(csv_path)
        constraints_path = os.path.join(TESTDATA_DIR, 'elements92rex.tdda')
        detectfile = os.path.join(self.tmp_dir,
                                  'elements118rex_detect_perc.csv')
        v = detect_df(df, constraints_path, report='fields',
                      outpath=detectfile, output_fields=['Z'],
                      per_constraint=True, rowindex_is_index=False)
        self.assertEqual(v.passes, 61)
        self.assertEqual(v.failures, 17)
        self.assertFileCorrect(detectfile, 'elements118rex_detect_perc.csv')

    def testDetectElements118rexToDataFrame(self):
        csv_path = os.path.join(TESTDATA_DIR, 'elements118.csv')
        df = pd.read_csv(csv_path)
        constraints_path = os.path.join(TESTDATA_DIR, 'elements92rex.tdda')
        v = detect_df(df, constraints_path, output_fields=['Z'],
                      rowindex_is_index=False)
        self.assertEqual(v.passes, 61)
        self.assertEqual(v.failures, 17)
        ddf = v.detected()
        self.assertStringCorrect(ddf.to_string(), 'elements118rex_detect.df')

    def testDetectElements118_csv_to_csv(self):
        self.detectElements('csv', 'csv')

    def testDetectElements118_csv_to_feather(self):
        self.detectElements('csv', 'feather')

    def testDetectElements118_feather_to_csv(self):
        self.detectElements('feather', 'csv')

    def testDetectElements118_feather_to_feather(self):
        self.detectElements('feather', 'feather')

    def detectElements(self, input, output):
        csv_path = os.path.join(TESTDATA_DIR, 'elements118.%s' % input)
        df = load_df(csv_path)
        constraints_path = os.path.join(TESTDATA_DIR, 'elements92.tdda')
        detect_name = 'elements118_detect_from_%s.%s' % (input, output)
        detectfile = os.path.join(self.tmp_dir, detect_name)
        v = detect_df(df, constraints_path, report='fields',
                      outpath=detectfile, output_fields=['Z'],
                      per_constraint=True, index=True,
                      rownumber_is_index=(input == 'feather'))
        self.assertEqual(v.detection.n_passing_records, 91)
        self.assertEqual(v.detection.n_failing_records, 27)
        if output == 'feather':
            # TODO: compare binary feather files and check they're the same
            pass
        else:
            self.assertFileCorrect(detectfile, detect_name)

    def testDetectDuplicates(self):
        iconstraints = FieldConstraints('i', [NoDuplicatesConstraint()])
        sconstraints = FieldConstraints('s', [NoDuplicatesConstraint()])
        constraints = DatasetConstraints([iconstraints, sconstraints])

        df1 = pd.DataFrame({'i': [1, 2, 3, 4, np.nan],
                            's': ['one', 'two', 'three', 'four', np.nan]})
        verifier1 = pdc.PandasConstraintVerifier(df1)
        v1 = verifier1.detect(constraints,
                              VerificationClass=pdc.PandasDetection)
        self.assertEqual(v1.passes, 2)
        self.assertEqual(v1.failures, 0)
        ddf1 = v1.detected()
        self.assertIsNone(ddf1)

        df2 = pd.DataFrame({'i': [1, 2, 3, 2, np.nan],
                            's': ['one', 'two', 'three', 'two', np.nan]})
        verifier2 = pdc.PandasConstraintVerifier(df2)
        v2 = verifier2.detect(constraints,
                              VerificationClass=pdc.PandasDetection,
                              per_constraint=True, output_fields=['i', 's'])
        self.assertEqual(v2.passes, 0)
        self.assertEqual(v2.failures, 2)
        ddf2 = v2.detected()
        self.assertStringCorrect(ddf2.to_string(), 'detect_dups.df')

    def testConstraintGenerationNoRex(self):
        self.constraintsGenerationTest(inc_rex=False)

    def testConstraintGenerationWithRex(self):
        self.constraintsGenerationTest(inc_rex=True)

    def constraintsGenerationTest(self, inc_rex=False):
        csv_path = os.path.join(TESTDATA_DIR, 'elements92.csv')
        df = pd.read_csv(csv_path)
        ref_name = 'elements92%s.tdda' % ('rex' if inc_rex else '')
        ref_constraints_path = os.path.join(TESTDATA_DIR, ref_name)
        with open(ref_constraints_path) as f:
            refjson = f.read()
        ref = native_definite(json.loads(refjson))
        constraints = discover_df(df, inc_rex=inc_rex)
        discovered = native_definite(json.loads(constraints.to_json()))
        discovered_fields = discovered['fields']
        ref_fields = ref['fields']
        self.assertEqual(set(discovered_fields.keys()),
                         set(ref_fields.keys()))
        for field, ref_field in ref_fields.items():
            ref_field = ref_fields[field]
            discovered_field = discovered_fields[field]
            self.assertEqual((field, set(discovered_field.keys())),
                             (field, set(ref_field.keys())))
            for c, expected in ref_field.items():
                actual = discovered_field[c]
                if type(expected) == float:
                    self.assertAlmostEqual(actual, expected, 4)
                elif type(expected) == list:
                    self.assertEqual(set(actual), set(expected))
                elif expected in ('int', 'real'):  # pandas too broken to
                                                   # get this right for now
                    self.assertTrue(actual in ('int', 'real'))
                else:
                    self.assertEqual(actual, expected)

    @unittest.skipIf(find_executable('tdda') is None, 'tdda not installed')
    def testTDDACommand(self):
        # similar test to testTDDACLI, but using the command-line wrapper
        self.command_line(wrapper=True)

    def testTDDACLI(self):
        # similar test to testTDDACommand, but using the console API
        self.command_line(wrapper=False)

    def testDDD_df(self):
        csv_path = os.path.join(TESTDATA_DIR, 'ddd.csv')
        df = pd.read_csv(csv_path)
        constraints_path = os.path.join(TESTDATA_DIR, 'ddd.tdda')
        v = verify_df(df, constraints_path)
        # expect 3 failures:
        #   - the pandas CSV reader will have read 'elevens' as an int
        #   - the pandas CSV reader will have read the date columns as strings
        self.assertEqual(v.passes, 58)
        self.assertEqual(v.failures, 3)

    def testDDD_csv(self):
        csv_path = os.path.join(TESTDATA_DIR, 'ddd.csv')
        constraints_path = os.path.join(TESTDATA_DIR, 'ddd.tdda')
        v = verify_df_from_file(csv_path, constraints_path, verbose=False)
        # expect 1 failure:
        #   - the enhanced CSV reader will have initially read 'elevens' as
        #     an int field and then (correctly) converted it to string, but
        #     it doesn't know that it would need to pad with initial zeros,
        #     so that means it will have computed its minimum as being '0'
        #     not '00', so the minimum string length won't be the same as
        #     Miro would compute (since Miro has the advantage of having
        #     additional metadata available when it read the CSV file, to
        #     tell it that 'elevens' is a string field.
        self.assertEqual(v.passes, 60)
        self.assertEqual(v.failures, 1)

    def testDDD_discover_and_verify(self):
        # both discovery and verification done using Pandas
        csv_path = os.path.join(TESTDATA_DIR, 'ddd.csv')
        c = discover_df_from_file(csv_path, constraints_path=None,
                                  verbose=False)
        tmpdir = tempfile.gettempdir()
        tmpfile = os.path.join(tmpdir, 'dddtestconstraints.tdda')
        with open(tmpfile, 'w') as f:
            f.write(c)
        v = verify_df_from_file(csv_path, tmpfile, report='fields',
                                verbose=False)
        self.assertEqual(v.passes, 61)
        self.assertEqual(v.failures, 0)

    def command_line(self, wrapper):
        # common code for testTDDACommand and testTDDACLI
        tmpdir = tempfile.gettempdir()
        constraintsdir = os.path.abspath(os.path.dirname(__file__))
        tddaTopDir = os.path.dirname(os.path.dirname(constraintsdir))
        dirs = ['referencetest_examples', 'constraints_examples',
                'rexpy_examples']
        testDataDir = os.path.join(tmpdir, 'constraints_examples', 'testdata')
        e92csv = os.path.join(testDataDir, 'elements92.csv')
        e118csv = os.path.join(testDataDir, 'elements118.csv')
        e92tdda = os.path.join(tmpdir, 'elements92.tdda')
        e118feather = os.path.join(testDataDir, 'elements118.feather')
        e92bads1 = os.path.join(tmpdir, 'elements92bads1.csv')
        e92bads2 = os.path.join(tmpdir, 'elements92bads2.csv')
        rmdirs(tmpdir, dirs)
        try:
            start = 'Copied example files for tdda.referencetest to'
            argv = ['tdda', 'examples', tmpdir]
            if wrapper:
                result = check_shell_output(argv)
                self.assertTrue(result.startswith(start))
                self.assertEqual(len(result.splitlines()), 3)
            else:
                main_with_argv(argv, verbose=False)
            argv = ['tdda', 'discover', e92csv, e92tdda]
            if wrapper:
                result = check_shell_output(argv)
                self.assertEqual(result, '')
            else:
                main_with_argv(argv, verbose=False)
            argv = ['tdda', 'verify', e92csv, e92tdda]
            self.assertTrue(os.path.exists(e92tdda))
            if wrapper:
                result = check_shell_output(argv)
            else:
                result = str(main_with_argv(argv, verbose=False))
            self.assertTrue(result.strip().endswith('SUMMARY:\n\n'
                                                    'Constraints passing: 72\n'
                                                    'Constraints failing: 0'))
            argv = ['tdda', 'verify', e118csv, e92tdda]
            if wrapper:
                result = check_shell_output(argv)
            else:
                result = str(main_with_argv(argv, verbose=False))
            self.assertTrue(result.strip().endswith('SUMMARY:\n\n'
                                                    'Constraints passing: 57\n'
                                                    'Constraints failing: 15'))
            argv = ['tdda', 'detect', e118csv, e92tdda, e92bads1,
                    '--per-constraint', '--output-fields', '--index']
            if wrapper:
                result = check_shell_output(argv)
            else:
                result = str(main_with_argv(argv, verbose=False))
            self.assertTrue(result.strip().endswith('SUMMARY:\n\n'
                                                    'Records passing: 91\n'
                                                    'Records failing: 27'))
            self.assertTrue(os.path.exists(e92bads1))
            self.assertFileCorrect(e92bads1, 'detect-els-cmdline.csv')
            argv = ['tdda', 'detect', e118feather, e92tdda, e92bads2,
                    '--per-constraint', '--output-fields', '--index']
            if wrapper:
                result = check_shell_output(argv)
            else:
                result = str(main_with_argv(argv, verbose=False))
            self.assertTrue(result.strip().endswith('SUMMARY:\n\n'
                                                    'Records passing: 91\n'
                                                    'Records failing: 27'))
            self.assertTrue(os.path.exists(e92bads2))
            self.assertFileCorrect(e92bads2, 'detect-els-cmdline2.csv')
        finally:
            rmdirs(tmpdir, dirs)


TestPandasConstraintVerifiers.set_default_data_location(TESTDATA_DIR)


def rmdirs(parent, dirs):
    for d in dirs:
        shutil.rmtree(os.path.join(parent, d), ignore_errors=True)


def check_shell_output(args):
    result = subprocess.check_output(UTF8DefiniteObject(args))
    return native_definite(result).replace('\r', '')


if __name__ == '__main__':
    ReferenceTestCase.main()




# -*- coding: utf-8 -*-

#
# Unit tests for reference data regeneration
#

from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import division

import os
import tempfile
import unittest

from tdda.referencetest.referencetest import ReferenceTest


class TestRegenerate(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.tmpdir = tempfile.gettempdir()

    def setUp(self):
        ReferenceTest.set_default_data_location(self.tmpdir)
        ReferenceTest.set_defaults(verbose=False)

    def tearDown(self):
        ReferenceTest.set_regeneration(regenerate=False)

    def test_regenerate_all(self):
        ReferenceTest.set_regeneration()
        ref = ReferenceTest(assert_fn=self.assertTrue)
        refname = 'regenerate_all.txt'
        reffile = os.path.join(self.tmpdir, refname)
        ref.assertStringCorrect('Start\nMiddle\nEnd\n', refname)
        with open(reffile) as f:
            self.assertEqual(f.read(), 'Start\nMiddle\nEnd\n')
        ref.assertStringCorrect('Completely different content', refname)
        with open(reffile) as f:
            self.assertEqual(f.read(), 'Completely different content')

    def test_regenerate_kinds(self):
        ReferenceTest.set_regeneration('csv')
        ref = ReferenceTest(assert_fn=self.assertTrue)
        txtname = 'regenerate.txt'
        csvname = 'regenerate.csv'
        txtfile = os.path.join(self.tmpdir, txtname)
        csvfile = os.path.join(self.tmpdir, csvname)
        with self.assertRaises(Exception):
            ref.assertStringCorrect('End\nMiddle\nStart\n', txtname,
                                     kind='txt')
        ref.assertStringCorrect('Start\nMiddle\nEnd\n', csvname, kind='csv')
        with open(csvfile) as f:
            self.assertEqual(f.read(), 'Start\nMiddle\nEnd\n')
        ref.assertStringCorrect('Completely different content', csvname,
                                kind='csv')
        with open(csvfile) as f:
            self.assertEqual(f.read(), 'Completely different content')


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

"""
Helpers for command-line option flags for discover and verify
"""

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import argparse
import sys


DISCOVER_HELP = '''
Optional flags are:

  * -r or --rex
      Include regular expression generation
  * -R or --norex
      Exclude regular expression generation (the default)
'''

VERIFY_HELP = '''
Optional flags are:

  * -a, --all
      Report all fields, even if there are no failures
  * -f, --fields
      Report only fields with failures
  * -7, --ascii
      Report in ASCII form, without using special characters.
  * --epsilon E
      Use this value of epsilon for fuzziness in comparing numeric values.
'''

DETECT_HELP = '''
Optional flags are:

  * -a, --all
      Report all fields, even if there are no failures
  * -f, --fields
      Report only fields with failures
  * -7, --ascii
      Report in ASCII form, without using special characters.
  * --epsilon E
      Use this value of epsilon for fuzziness in comparing numeric values.
  * --write-all
      Include passing records in the output.
  * --per-constraint
      Write one column per failing constraint, as well as the n_failures
      total column for each row.
  * --output-fields FIELD1 FIELD2 ...
      Specify original columns to write out.
      If used with no field names, all original columns will be included.
  * --index
      Include a row-number index in the output file.
      The row number is automatically included if no output fields are
      specified. Rows are usually numbered from 1, unless the (feather)
      input file already has an index.

'''

def discover_parser(usage=''):
    formatter = argparse.RawDescriptionHelpFormatter
    parser = argparse.ArgumentParser(prog='tdda discover',
                                     epilog=usage + DISCOVER_HELP,
                                     formatter_class=formatter)
    parser.add_argument('-?', '--?', action='help',
                        help='same as -h or --help')
    parser.add_argument('-r', '--rex', action='store_true',
                        help='include regular expression generation')
    parser.add_argument('-R', '--norex', action='store_true',
                        help='exclude regular expression generation')
    return parser


def discover_flags(parser, args, params):
    flags, more = parser.parse_known_args(args)
    if len(more) > 0:
        print(parser.epilog, file=sys.stderr)
        sys.exit(1)
    params['inc_rex'] = flags.rex
    return flags


def verify_parser(usage=''):
    formatter = argparse.RawDescriptionHelpFormatter
    parser = argparse.ArgumentParser(prog='tdda verify',
                                     epilog=usage + VERIFY_HELP,
                                     formatter_class=formatter)
    parser.add_argument('-?', '--?', action='help',
                        help='same as -h or --help')
    parser.add_argument('-a', '--all', action='store_true',
                        help='report all fields, even if there are '
                             'no failures')
    parser.add_argument('-f', '--fields', action='store_true',
                        help='report only fields with failures')
    parser.add_argument('-7', '--ascii', action='store_true',
                        help='report without using special characters')
    parser.add_argument('-type_checking', action='store_true',
                        help='strict or sloppy')
    parser.add_argument('-epsilon', '--epsilon', nargs=1,
                        help='epsilon fuzziness')
    return parser


def detect_parser(usage=''):
    formatter = argparse.RawDescriptionHelpFormatter
    parser = argparse.ArgumentParser(prog='tdda detect',
                                     epilog=usage + DETECT_HELP,
                                     formatter_class=formatter)
    parser.add_argument('-?', '--?', action='help',
                        help='same as -h or --help')
    parser.add_argument('-a', '--all', action='store_true',
                        help='report all fields, even if there are '
                             'no failures')
    parser.add_argument('-f', '--fields', action='store_true',
                        help='report only fields with failures')
    parser.add_argument('-7', '--ascii', action='store_true',
                        help='report without using special characters')
    parser.add_argument('-type_checking', action='store_true',
                        help='strict or sloppy')
    parser.add_argument('-epsilon', '--epsilon', nargs=1,
                        help='epsilon fuzziness')
    parser.add_argument('--write-all', action='store_true',
                        help='Include passing records')
    parser.add_argument('--per-constraint', action='store_true',
                        help='Write one column per failing constraint '
                             'in addition to n_failures')
    parser.add_argument('--output-fields', nargs='*',
                        help='Specify original columns to write out. '
                             'If used with no field names, then '
                             'all original columns will be included')
    parser.add_argument('--index', action='store_true',
                        help='Include a row-number index in the output file '
                             'when detecting. Rows are usually numbered from '
                             '1, unless the input file already has an index.')
    parser.add_argument('--int', dest='boolean_ints', action='store_true',
                        help='Write out boolean fields as integers, with '
                             '1 for true and 0 for false.')
    return parser


def verify_flags(parser, args, params):
    flags, more = parser.parse_known_args(args)
    if len(more) > 0:
        print(parser.epilog, file=sys.stderr)
        sys.exit(1)
    params.update({
        'report': 'all',
        'ascii': False,
    })
    if flags.all:
        params['report'] = 'all'
    elif flags.fields:
        params['report'] = 'fields'
    if flags.ascii:
        params['ascii'] = True
    if flags.type_checking:
        params['type_checking'] = True
    if flags.epsilon:
        params['epsilon'] = float(flags.epsilon[0])
    return flags


def detect_flags(parser, args, params):
    flags, more = parser.parse_known_args(args)
    if len(more) > 0:
        print(parser.epilog, file=sys.stderr)
        sys.exit(1)
    params.update({
        'report': 'records',
        'ascii': False,
    })
    if flags.all:
        params['report'] = 'all'
    elif flags.fields:
        params['report'] = 'records'
    if flags.ascii:
        params['ascii'] = True
    if flags.type_checking:
        params['type_checking'] = True
    if flags.epsilon:
        params['epsilon'] = float(flags.epsilon[0])
    if flags.write_all:
        params['write_all'] = True
    if flags.per_constraint:
        params['per_constraint'] = True
    if flags.index:
        params['index'] = True
    if flags.boolean_ints:
        params['boolean_ints'] = True
    if flags.output_fields is not None:
        params['output_fields'] = flags.output_fields
    params['in_place'] = False  # Only applicable in API case
    params['report'] = 'records'
    return flags




# -*- coding: utf-8 -*-

"""
This provides all of the methods in the
:py:class:`~tdda.referencetest.referencetest.ReferenceTest` class,
in a way that allows them to be used as ``pytest`` fixtures.

This allows these functions to be called from tests running from the
``pytest`` framework.

For example::

    from tdda.referencetest import referencepytest
    import my_module

    def test_my_csv_function(ref):
        resultfile = my_module.my_csv_function(ref.tmp_dir)
        ref.assertCSVFileCorrect(resultfile, 'result.csv')

    def test_my_pandas_dataframe_function(ref):
        resultframe = my_module.my_dataframe_function()
        ref.assertDataFrameCorrect(resultframe, 'result.csv')

    def test_my_table_function(ref):
        result = my_module.my_table_function()
        ref.assertStringCorrect(result, 'table.txt', kind='table')

    def test_my_graph_function(ref):
        result = my_module.my_graph_function()
        ref.assertStringCorrect(result, 'graph.txt', kind='graph')

    class TestMyClass:
        def test_my_other_table_function(ref):
            result = my_module.my_other_table_function()
            ref.assertStringCorrect(result, 'table.txt', kind='table')

with a ``conftest.py`` containing::

    import pytest
    from tdda.referencetest import referencepytest

    def pytest_addoption(parser):
        referencepytest.addoption(parser)

    def pytest_collection_modifyitems(session, config, items):
        referencepytest.tagged(config, items)

    @pytest.fixture(scope='module')
    def ref(request):
        r = referencepytest.ref(request)
        r.set_data_location('testdata')
        return r

Tagged Tests
~~~~~~~~~~~~

If the tests are run with the ``--tagged``
command-line option, then only tests that have been decorated with
``referencetest.tag``, are run. This is a mechanism for allowing
only a chosen subset of tests to be run, which is useful during
development. The ``@tag`` decorator can be applied to test functions,
test classes and test methods.

If the tests are run with the ``--istagged`` command-line option,
then no tests are run; instead, the
framework reports the full module names of any test classes or functions
that have been decorated with ``@tag``, or classes which contain any
test methods that have been decorated with ``@tag``.

For example::

    from tdda.referencetest import referencepytest, tag
    import my_module

    @tag
    def test_a(ref):
        ...

    def test_b(ref):
        ...

    @tag
    class TestMyClass:
        def test_x(self):
            ...

        def test_y(self):
            ...

If run with ``pytest --tagged``, only the tagged tests are
run (``test_a``, ``TestMyClass.test_x`` and ``TestMyClass.test_y``).

Regeneration of Results
~~~~~~~~~~~~~~~~~~~~~~~

When ``pytest`` is run with ``--write-all`` or ``--write``, it causes
the framework to regenerate reference data files. Different kinds of
reference results can be regenerated by passing in a comma-separated list
of *kind* names immediately after the ``--write`` option. If no list
of *kind* names is provided, then all test results will be regenerated.

If the **-s** option is also provided (to disable ``pytest``
output capturing), it will report the names of all the files it has
regenerated.

To regenerate all reference results (or generate them for the first time)

.. code-block:: bash

   pytest -s --write-all
    
To regenerate just a particular kind of reference (e.g. table results)

.. code-block:: bash

    pytest -s --write table

To regenerate a number of different kinds of reference (e.g. both table
and graph results)

.. code-block:: bash

    pytest -s --write table graph

``pytest`` Integration Details
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

In addition to all of the methods from
:py:class:`~tdda.referencetest.referencetest.ReferenceTest`,
the following functions are provided, to allow easier integration
with the ``pytest`` framework.
"""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import sys

import pytest

from tdda.referencetest.referencetest import ReferenceTest, tag


def pytest_assert(x, msg):
    # assertion using standard python assert statement, as expected by pytest.
    assert x, msg


def ref(request):
    """
    Support for dependency injection via a ``pytest`` fixture.

    A test's conftest.py should define a fixture function for injecting
    a :py:class:`~tdda.referencetest.referencetest.ReferenceTest` instance,
    which should just call this function.

    This allows tests to get access to a private instance of that class.
    """
    if request.config.getoption('--wquiet'):
        ReferenceTest.set_defaults(verbose=False)
    if request.config.getoption('--write-all'):
        ReferenceTest.set_regeneration()
    else:
        regen = request.config.getoption('--write')
        if regen:
            for r in regen:
                for kind in r.split(','):
                    ReferenceTest.set_regeneration(kind)
    return ReferenceTest(pytest_assert)


def set_defaults(**kwargs):
    """
    This provides a mechanism for setting default attributes in
    the :py:class:`~tdda.referencetest.referencetest.ReferenceTest` class.

    It takes the same parameters as
    :py:meth:`tdda.referencetest.referencetest.ReferenceTest.set_defaults`,
    and can be used for setting parameters such as the *tmp_dir* property.

    If you want the same defaults for all your tests, it can be easier to
    set them with a call to this function, rather than having to set them
    explicitly in each test (or in your ``@pytest.fixture`` *ref* definition
    in your ``conftest.py`` file).
    """
    ReferenceTest.set_defaults(**kwargs)


def set_default_data_location(location, kind=None):
    """
    This provides a mechanism for setting the default reference data
    location in the :py:class:`~tdda.referencetest.referencetest.ReferenceTest`
    class.

    It takes the same parameters as
    :py:meth:`tdda.referencetest.referencetest.ReferenceTest.set_default_data_location`.

    If you want the same data locations for all your tests, it can be easier
    to set them with calls to this function, rather than having to set them
    explicitly in each test (or using
    :py:meth:`~tdda.referencetest.referencetest.ReferenceTest.set_data_location`
    in your ``@pytest.fixture`` *ref* definition in your ``conftest.py`` file).
    """
    ReferenceTest.set_default_data_location(location, kind=kind)


def addoption(parser):
    """
    Support for the --write and --write-all command-line options.

    A test's ``conftest.py`` file should declare extra options by
    defining a ``pytest_addoption`` function which should just call this.

    It extends pytest to include **--write** and **--write-all** option
    flags which can be used to control regeneration of reference results.
    """
    try:
        parser.addoption('--write', action='store', nargs='+', default=None,
                         help='--write: rewrite named reference results kinds')
        parser.addoption('--write-all', action='store_true',
                         help='--write-all: rewrite all reference results')
        parser.addoption('--wquiet', action='store_true',
                         help='--wquiet: when rewriting results, '
                              'do so quietly')
        parser.addoption('--tagged', action='store_true',
                         help='--tagged: only run tagged tests')
        parser.addoption('--istagged', action='store_true',
                         help='--istagged: report tagged tests, '
                              'without running')
    except ValueError:
        # ignore attempts to add parser options multiple times
        pass


def tagged(config, items):
    """
    Support for @tag to mark tests to be run with --tagged or reported
    with --istagged.

    A test's ``conftest.py`` file should declare tags by defining a
    ``pytest_collection_modifyitems`` function which should just call this.

    It extends pytest to recognize the ``--tagged`` and ``--istagged``
    command-line flags, to restrict testing to tagged tests only.
    """
    runtagged = config.getoption('--tagged')
    showtagged = config.getoption('--istagged')
    shownclasses = set()
    if runtagged or showtagged:
        if showtagged:
            print()
        for f in list(items):
            tagged = False
            cls = None
            if hasattr(f.obj, '__self__'):
                cls = f.obj.__self__.__class__
                tagged = getattr(cls, '_tagged', None)
            if not tagged:
                tagged = getattr(f.obj, '_tagged', None)
            if showtagged or not tagged:
                items.remove(f)
            if tagged and showtagged:
                if cls:
                    if cls not in shownclasses:
                        print('%s.%s' % (f.obj.__module__, cls.__name__))
                    shownclasses.add(cls)
                else:
                    print('%s.%s' % (f.obj.__module__, f.name))

